"""
Currencies tests
"""

from datetime import date
from django.test import TestCase
from iso4217 import Currency as Iso4217
from rest_framework import status
from rest_framework.test import APIClient

from .models import Currency


class CurrencyTestCase(TestCase):
    """
    Tests for Currency class
    """

    def test_all_currencies(self):
        """
        Test listing of all currencies
        """
        self.assertEqual(len(list(Currency.all_currencies())), len(Iso4217))

    def test_ordered_all_currencies(self):
        """
        Test sorting of list of currencies
        """
        self.assertEqual(Currency.all_currencies(ordering='code')[0].code, 'AED')
        self.assertEqual(Currency.all_currencies(ordering='name')[0].code, 'AED')
        self.assertEqual(Currency.all_currencies(ordering='currency_name')[0].code, 'AFN')
        self.assertEqual(Currency.all_currencies(ordering='exponent')[0].code, 'XOF')
        self.assertEqual(Currency.all_currencies(ordering='number')[0].code, 'ALL')
        self.assertEqual(Currency.all_currencies(ordering='value')[0].code, 'AED')

    def test_creation(self):
        """
        Test creation of a currency
        """
        c = Currency('EUR')
        self.assertEqual(c.name, 'eur')
        self.assertEqual(c.value, 'EUR')
        self.assertEqual(c.currency_name, 'Euro')
        self.assertEqual(c.number, 978)
        self.assertEqual(c.code, 'EUR')
        self.assertEqual(c.symbol, '€')

    def test_is_valid(self):
        """
         Test currency validation
        """
        self.assertTrue(Currency.is_valid('EUR'))
        self.assertFalse(Currency.is_valid('REU'))

    def test_get_for_country(self):
        """
        Test getting currency for a country
        """
        currencies = Currency.get_for_country('FR')
        self.assertIsNotNone(currencies)
        c = currencies[0]
        self.assertEqual(c.name, 'eur')
        self.assertEqual(c.value, 'EUR')
        self.assertEqual(c.currency_name, 'Euro')
        self.assertEqual(c.number, 978)
        self.assertEqual(c.code, 'EUR')

    def test_countries(self):
        """
        Test list of countrries for a currency
        """
        c = Currency('EUR')
        self.assertIsNotNone(c.countries)

    def test_get_rates(self):
        """
        Test getting rates for a currency
        """
        c = Currency('EUR')
        rates = c.get_rates()
        self.assertIsNotNone(rates)


class CurrencyAPITestCase(TestCase):
    """
    Tests for Currency API
    """

    def test_list_request(self):
        """
        Test list of currencies
        """
        client = APIClient()
        response = client.get('/currencies/', format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), len(Currency.all_currencies()))
        self.assertEqual(response.data[0].get('code'), 'AED')

    def test_list_ordered_request(self):
        """
        Test list with sorting
        """
        client = APIClient()
        response = client.get('/currencies/', data={'ordering': 'code'}, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.data), len(Currency.all_currencies()))
        self.assertEqual(response.data[0].get('code'), 'AED')

    def test_retrieve_request(self):
        """
        Test retrieval of a currency
        """
        client = APIClient()
        response = client.get('/currencies/EUR/', format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data.get('code'), 'EUR')
        self.assertEqual(response.data.get('symbol'), '€')

    def test_get_countries_request(self):
        """
        Test listing countries for a currency
        """
        client = APIClient()
        response = client.get('/currencies/AFN/countries/', format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data[0].get('alpha_2'), 'AF')

    def test_get_rates_request(self):
        """
        Test listing rates for a currency (use /rates/ instead)
        """
        client = APIClient()
        response = client.get('/currencies/AFN/rates/', format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_rates_with_base_request(self):
        """
        Test listing rates for a base currency (use /rates/ instead)
        """
        client = APIClient()
        response = client.get(
            '/currencies/AFN/rates/',
            data={
                'base_currency': 'USD'
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_rates_with_start_date_request(self):
        """
        Test listing rates for a currency (use /rates/ instead) from a date
        """
        client = APIClient()
        response = client.get(
            '/currencies/AFN/rates/',
            data={
                'start_date': date.today().strftime('%Y-%m-%d')
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_get_rates_with_end_date_request(self):
        """
        Test listing rates for a currency (use /rates/ instead) with end date
        """
        client = APIClient()
        response = client.get(
            '/currencies/AFN/rates/',
            data={
                'end_date': date.today().strftime('%Y-%m-%d')
            },
            format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)



# Generated by Django 3.1.2 on 2020-10-19 07:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('units', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customunit',
            name='key',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=255, null=True),
        ),
    ]



#!/usr/bin/env python
# https://github.com/datamafia/color-proximity
"""color-proximity module."""

from __future__ import division

from math import sqrt, pow as poww


def hextorgb(hex):
    c = hex
    r, g, b = 0, 0, 0
    if '#' in hex:
        c = hex[1:]
    if len(c) == 8:
        # alpha channel
        c = c[2:]
    if len(c) == 3:
        # simplified hex
        c = c[0] * 2 + c[1] * 2 + c[2] * 2
    try:
        r = int(c[:2], 16)
        g = int(c[2:4], 16)
        b = int(c[4:], 16)
    except (ValueError, TypeError):
        return 0, 0, 0
    return r, g, b


class ColorProximity(object):
    """Color Proximity."""

    def proximity(self, a_set, b_set):
        """Return proximity value of a pair RGB tuple or list
        Init, convert list/tupel of int,int,int to color proximity value
        Args:
        a : required, list or tuple containting int, 0-255 RGB color value
        b : required, list or tuple containting int, 0-255 RGB color value
        Returns :
        float
        """
        # support for RBG+alpha, slice to ignore alpha
        if len(a_set) > 3:
            a_set = a_set[:3]
        if len(b_set) > 3:
            b_set = b_set[:3]
        c_1 = self.rgb2lab(a_set)  # int int int
        c_2 = self.rgb2lab(b_set)

        return sqrt(poww(c_1[0] - c_2[0], 2) + poww(c_1[1] - c_2[1], 2) + poww(c_1[2] - c_2[2], 2))

    @classmethod
    def rgb2lab(cls, c_set):
        """RGB to LAB conversion"""

        # """RGB to L.a.b. conversion
        # Transform (int, int, int) from RGB to XYZ (CIE1931) to L.a.b.
        #    http://en.wikipedia.org/wiki/CIE_1931_color_space
        #    http://en.wikipedia.org/wiki/Lab_color_space
        # Args:
        #    c : required, list or tuple of 3 int (RGB)
        # Return :
        #    Lab value as 3 element list
        # """

        rgb = [0, 0, 0]
        xyz = [0, 0, 0]
        lab = [0, 0, 0]

        i = 0
        while len(c_set) > i:
            value = float(c_set[i] / 255)
            if value > 0.04045:
                value = poww(((value + 0.055) / 1.055), 2.4)
            else:
                value = value / 12.92
            rgb[i] = value * 100
            i += 1

        xyz[0] = (rgb[0] * 0.4124 + rgb[1] * 0.3576 + rgb[2] * 0.1805) / 95.047
        # // ref_X =  95.047   Observer= 2deg, Illuminant= D65
        xyz[1] = (rgb[0] * 0.2126 + rgb[1] * 0.7152 + rgb[2] * 0.0722) / 100.0
        # // ref_Y = 100.000
        xyz[2] = (rgb[0] * 0.0193 + rgb[1] * 0.1192 + rgb[2] * 0.9505) / 108.883
        # // ref_Z = 108.883

        i = 0
        while len(xyz) > i:
            value = xyz[i]
            if value > 0.008856:
                value = poww(value, 1 / 3)
            else:
                value = (7.787 * value) + (16 / 116)
            xyz[i] = value
            i += 1

        lab[0] = round(((116 * xyz[1]) - 16), 3)
        lab[1] = round((500 * (xyz[0] - xyz[1])), 3)
        lab[2] = round((200 * (xyz[1] - xyz[2])), 3)

        return lab



"""
Units module
"""

from django.utils.translation import ugettext_lazy as _, pgettext_lazy

DIMENSIONS = {
    '[length]': {'name': _('length'), 'dimension': 'meter', 'symbol': 'L'},
    '[time]': {'name': _('time'), 'dimension': 'second', 'symbol': 'T'},
    '[current]': {'name': _('current'), 'dimension': 'ampere', 'symbol': 'I'},
    '[luminosity]': {'name': _('luminosity'), 'dimension': 'candela * radian ** 2',
                     'symbol': 'J'},
    '[mass]': {'name': _('mass'), 'dimension': 'kilogram', 'symbol': 'M'},
    '[substance]': {'name': _('substance'), 'dimension': 'mole', 'symbol': 'N'},
    '[temperature]': {'name': _('temperature'), 'dimension': 'kelvin', 'symbol': 'Θ'},
    '[]': {'name': _('constant'), 'dimension': 'bit', 'symbol': 'b'},
    '[area]': {'name': _('area'), 'dimension': 'meter ** 2', 'symbol': 'L²'},
    '[volume]': {'name': _('volume'), 'dimension': 'meter ** 3', 'symbol': 'L³'},
    '[frequency]': {'name': _('frequency'), 'dimension': 'count / second',
                    'symbol': 'T⁻¹'},
    '[wavenumber]': {'name': _('wavenumber'), 'dimension': '1 / meter', 'symbol': 'L⁻¹'},
    '[velocity]': {'name': _('velocity'), 'dimension': 'meter / second',
                   'symbol': 'L.T⁻¹'},
    '[acceleration]': {'name': _('acceleration'), 'dimension': 'meter / second ** 2',
                       'symbol': 'L.T⁻²'},
    '[force]': {'name': _('force'), 'dimension': 'kilogram * meter / second ** 2',
                'symbol': 'F'},
    '[energy]': {'name': _('energy'), 'dimension': 'kilogram * meter ** 2 / second ** 2',
                 'symbol': 'E'},
    '[power]': {'name': _('power'), 'dimension': 'kilogram * meter ** 2 / second ** 3',
                'symbol': 'P'},
    '[density]': {'name': _('density'), 'dimension': 'kilogram / meter ** 3',
                  'symbol': 'M.L⁻³'},
    '[pressure]': {'name': _('pressure'), 'dimension': 'kilogram / meter / second ** 2',
                   'symbol': 'Pa'},
    '[torque]': {'name': _('torque'), 'dimension': 'kilogram * meter ** 2 / second ** 2',
                 'symbol': 'F.L'},
    '[viscosity]': {'name': _('viscosity'), 'dimension': 'kilogram / meter / second',
                    'symbol': 'M.L⁻¹.T⁻¹'},
    '[kinematic_viscosity]': {'name': _('kinematic viscosity'),
                              'dimension': 'meter ** 2 / second',
                              'symbol': 'L².T⁻¹'},
    '[fluidity]': {'name': _('fluidity'), 'dimension': 'meter * second / kilogram',
                   'symbol': 'L.T.M⁻¹'},
    '[concentration]': {'name': _('concentration'), 'dimension': 'mole / meter ** 3',
                        'symbol': 'N.L⁻³'},
    '[activity]': {'name': _('activity'), 'dimension': 'mole / second',
                   'symbol': 'N.T⁻¹'},
    '[entropy]': {'name': _('entropy'),
                  'dimension': 'kilogram * meter ** 2 / kelvin / second ** 2',
                  'symbol': ''},
    '[molar_entropy]': {'name': _('molar entropy'),
                        'dimension': 'kilogram * meter ** 2 / kelvin / mole / second ** 2',
                        'symbol': 'M.L².Θ⁻¹.N⁻¹.T⁻²'
                        },
    '[heat_transmission]': {'name': _('heat transmission'),
                            'dimension': 'kilogram / second ** 2',
                            'symbol': 'M.T⁻²'
                            },
    '[luminance]': {'name': _('luminance'),
                    'dimension': 'candela * radian ** 2 / meter ** 2',
                    'symbol': 'J.rad².L⁻²'
                    },
    '[luminous_flux]': {'name': _('luminous flux'),
                        'dimension': 'candela * radian ** 2', 'symbol': 'J.rad²'
                        },
    '[illuminance]': {'name': _('illuminance'),
                      'dimension': 'candela * radian ** 2 / meter ** 2',
                      'symbol': 'J.rad².L⁻²'
                      },
    '[intensity]': {'name': _('intensity'), 'dimension': 'kilogram / second ** 3',
                    'symbol': 'M.T⁻³'},
    '[charge]': {'name': _('charge'), 'dimension': 'ampere * second', 'symbol': 'Q'},
    '[electric_potential]': {'name': _('electric potential'),
                             'dimension': 'kilogram * meter ** 2 / ampere / second ** 3',
                             'symbol': 'V'
                             },
    '[electric_field]': {'name': _('electric field'),
                         'dimension': 'kilogram * meter / ampere / second ** 3',
                         'symbol': ''
                         },
    '[resistance]': {'name': _('resistance'),
                     'dimension': 'kilogram * meter ** 2 / ampere ** 2 / second ** 3',
                     'symbol': 'Ω'
                     },
    '[conductance]': {'name': _('conductance'),
                      'dimension': 'ampere ** 2 * second ** 3 / kilogram / meter ** 2',
                      'symbol': 'S'
                      },
    '[capacitance]': {'name': _('capacitance'),
                      'dimension': 'ampere ** 2 * second ** 4 / kilogram / meter ** 2',
                      'symbol': 'F'
                      },
    '[inductance]': {'name': _('inductance'),
                     'dimension': 'kilogram * meter ** 2 / ampere ** 2 / second ** 2',
                     'symbol': 'H'
                     },
    '[magnetic_flux]': {'name': _('magnetic flux'),
                        'dimension': 'kilogram * meter ** 2 / ampere / second ** 2',
                        'symbol': 'Wb'
                        },
    '[magnetic_field]': {'name': _('magnetic field'),
                         'dimension': 'kilogram / ampere / second ** 2',
                         'symbol': ''
                         },
    '[magnetomotive_force]': {'name': _('magnetomotive force'), 'dimension': 'ampere',
                              'symbol': 'I'
                              },
    '[electric_dipole]': {'name': _('electric dipole'),
                          'dimension': 'ampere * meter * second',
                          'symbol': ''
                          },
    '[electric_quadrupole]': {'name': _('electric quadrupole'),
                              'dimension': 'ampere * meter ** 2 * second',
                              'symbol': ''
                              },
    '[magnetic_dipole]': {'name': _('magnetic dipole'),
                          'dimension': 'ampere * meter ** 2', 'symbol': ''
                          },
    '[printing_unit]': {'name': _('printing unit'), 'dimension': 'pixel',
                        'symbol': 'pixel'
                        },
    '[gaussian_charge]': {'name': _('gaussian charge'),
                          'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second',
                          'symbol': ''
                          },
    '[gaussian_current]': {'name': _('gaussian current'),
                           'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second ** 2',
                           'symbol': ''
                           },
    '[gaussian_electric_potential]': {'name': _('gaussian electric potential'),
                                      'dimension': 'kilogram ** 0.5 * meter ** 0.5 / second',
                                      'symbol': ''
                                      },
    '[gaussian_electric_field]': {'name': _('gaussian electric field'),
                                  'dimension': 'kilogram ** 0.5 / meter ** 0.5 / second',
                                  'symbol': ''
                                  },
    '[gaussian_electric_displacement_field]': {
        'name': _('gaussian electric displacement field'),
        'dimension': 'kilogram ** 0.5 / meter ** 0.5 / second',
        'symbol': ''
    },
    '[gaussian_electric_flux]': {'name': _('gaussian electric flux'),
                                 'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second',
                                 'symbol': ''
                                 },
    '[gaussian_magnetic_field]': {'name': _('gaussian magnetic field'),
                                  'dimension': 'kilogram ** 0.5 / meter ** 0.5 / second',
                                  'symbol': ''
                                  },
    '[gaussian_magnetic_field_strength]': {'name': _('gaussian magnetic field strength'),
                                           'dimension': 'kilogram ** 0.5 / meter ** 0.5 / second',
                                           'symbol': ''
                                           },
    '[gaussian_magnetic_flux]': {'name': _('gaussian magnetic flux'),
                                 'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second',
                                 'symbol': ''
                                 },
    '[gaussian_resistance]': {'name': _('gaussian resistance'),
                              'dimension': 'second / meter',
                              'symbol': ''
                              },
    '[gaussian_resistivity]': {'name': _('gaussian resistivity'), 'dimension': 'second',
                               'symbol': ''
                               },
    '[gaussian_capacitance]': {'name': _('gaussian capacitance'), 'dimension': 'meter',
                               'symbol': ''
                               },
    '[gaussian_inductance]': {'name': _('gaussian inductance'),
                              'dimension': 'second ** 2 / meter',
                              'symbol': ''
                              },
    '[gaussian_conductance]': {'name': _('gaussian conductance'),
                               'dimension': 'meter / second',
                               'symbol': ''
                               },
    '[esu_charge]': {'name': _('esu charge'),
                     'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second',
                     'symbol': ''
                     },
    '[esu_current]': {'name': _('esu current'),
                      'dimension': 'kilogram ** 0.5 * meter ** 1.5 / second ** 2',
                      'symbol': ''
                      },
    '[esu_electric_potential]': {'name': _('esu electric potential'),
                                 'dimension': 'kilogram ** 0.5 * meter ** 0.5 / second',
                                 'symbol': ''
                                 },
    '[esu_magnetic_flux]': {'name': _('esu magnetic flux'),
                            'dimension': 'kilogram ** 0.5 * meter ** 0.5',
                            'symbol': ''
                            },
    '[esu_magnetic_field]': {'name': _('esu magnetic field'),
                             'dimension': 'kilogram ** 0.5 / meter ** 1.5',
                             'symbol': ''
                             },
    '[compounded]': {'name': _('Compound dimension'),
                     'dimension': 'undefined',
                     'symbol': ''
                },
    '[custom]': {'name': _('Custom dimension'), 'dimension': 'undefined', 'symbol': ''},
}

UNIT_SYSTEM_BASE_AND_DERIVED_UNITS = {
    'SI': {
        '[]': 'count',
        '[time]': 'second',
        '[length]': 'meter',
        '[mass]': 'kilogram',
        '[current]': 'ampere',
        '[temperature]': 'kelvin',
        '[substance]': 'mole',
        '[luminosity]': 'candela',
        '[area]': 'square_meter',
        '[volume]': 'cubic_meter',
        '[frequency]': 'meter_per_second',
        '[acceleration]': 'meter_per_square_second',
        '[force]': 'newton',
        '[energy]': 'joule',
        '[electric_charge]': 'coulomb',
        '[electric_potential]': 'volt',
        '[capacitance]': 'farad',
        '[resistance]': 'ohm',
        '[conductance]': 'siemens',
        '[magnetic_flux]': 'weber',
        '[inductance]': 'henry',
        '[luminous_flux]': 'lumen',
        '[illuminance]': 'lux',
        '[activity]': 'katal',
    },
    'Planck': {
        '[length]': 'planck_length',
        '[mass]': 'planck_mass',
        '[time]': 'planck_time',
        '[temperature]': 'planck_temperature',
        '[current]': 'planck_current'
    },
    'US': {
        '[length]': 'yard',
        '[mass]': 'pound'
    },
    'atomic': {
        '[length]': 'bohr',
        '[mass]': 'electron_mass',
        '[time]': 'atomic_unit_of_time',
        '[current]': 'atomic_unit_of_current',
        '[temperature]': 'atomic_unit_of_temperature'
    },
    'cgs': {
        '[length]': 'centimeter',
        '[mass]': 'gram',
        '[time]': 'second',
        '[velocity]': 'centimeter_per_second',
        '[acceleration]': 'galileo',
        '[force]': 'dyn',
        '[energy]': 'erg',
        '[power]': 'watt',
        '[pressure]': 'barye',
        '[dynamic_viscosity]': 'poise',
        '[kinematic_viscosity]': 'stokes',
        '[wavenumber]': 'reciprocal_centimeter'
    },
    'imperial': {
        '[length]': 'yard',
        '[mass]': 'pound'
    },
    'mks': {
        '[acceleration]': 'meter_per_square_second',
        '[capacitance]': 'farad',
        '[charge]': 'coulomb',
        '[current]': 'ampere',
        '[electric_field]': 'volt_per_meter',
        '[electric_potential]': 'volt',
        '[energy]': 'joule',
        '[power]': 'watt',
        '[force]': 'newton',
        '[inductance]': 'henry',
        '[length]': 'meter',
        '[magnetic_field]': 'tesla',
        '[magnetic_flux]': 'weber',
        '[mass]': 'kilogram',
        '[pressure]': 'pascal',
        '[resistance]': 'ohm',
        '[temperature]': 'kelvin',
        '[time]': 'second',
        '[velocity]': 'meter_per_second'
    }
}

PREFIX_SYMBOL = {
    'yotta': 'Y',
    'zetta': 'Z',
    'exa': 'E',
    'peta': 'P',
    'tera': 'T',
    'giga': 'G',
    'mega': 'M',
    'kilo': 'k',
    'hecto': 'h',
    'deca': 'da',
    'deci': 'd',
    'centi': 'c',
    'milli': 'm',
    'micro': 'μ',
    'nano': 'n',
    'pico': 'p',
    'femto': 'f',
    'atto': 'a',
    'zepto': 'z',
    'yocto': 'y'
}

ADDITIONAL_BASE_UNITS = {
    'SI': {
        'meter_per_second': {
            'name': _('meter per second'),
            'symbol': 'm.s⁻¹',
            'relation': '1 meter / second'
        },
        'meter_per_square_second': {
            'name': _('meter per square second'),
            'symbol': 'm.s⁻²',
            'relation': '1 meter / second ** 2'
        },
        'square_meter': {
            'name': _('square meter'),
            'symbol': 'm²',
            'relation': '1 meter ** 2'
        },
        'cubic_meter': {
            'name': _('cubic meter'),
            'symbol': 'm³',
            'relation': '1 meter ** 3'
        },
    }
}

### Dictionnary with translations
UNIT_EXTENDED_DEFINITION = {
    'K_alpha_Cu_d_220': {
        'name': _('Copper Kα'),
        'symbol': 'Cu Kα',
    },
    'K_alpha_Mo_d_220': {
        'name': _('Molybdenum K-α'),
        'symbol': 'MO Kα',
    },
    'K_alpha_W_d_220': {
        'name': _('Tungsten K-α'),
        'symbol': _('W Kα'),
    },
    'RKM': {
        'name': _('RKM'),
        'symbol': 'RKM',
        'obsolete': True,
    },
    'UK_force_ton': {
        'name': _('long ton-force'),
        'symbol': 'long ton-force',
        'obsolete': True,
    },
    'UK_hundredweight': {
        'name': 'long hundredweight',
        'symbol': 'long hundredweight',
        'obsolete': True,
    },
    'UK_ton': {
        'name': _('long ton'),
        'symbol': 'long ton',
        'obsolete': True,
    },
    'US_force_ton': {
        'name': _('short ton force'),
        'symbol': 'short ton-force',
        'obsolete': True,
    },
    'US_hundredweight': {
        'name': _('short hundredweight'),
        'symbol': 'short hundredweight',
        'obsolete': True,
    },
    'US_ton': {
        'name': _('short ton'),
        'symbol': 'short ton',
        'obsolete': True,
    },
    'US_international_ampere': {
        'name': _('US International Ampere'),
        'symbol': 'US international A',
        'obsolete': True,
    },
    'US_international_ohm': {
        'name': _('US international Ohm'),
        'symbol': 'US international Ω',
        'obsolete': True,
    },
    'US_international_volt': {
        'name': _('US international Volt'),
        'symbol': 'US international V',
        'obsolete': True,
    },
    'US_therm': {
        'name': _('US therm'),
        'symbol': 'thm',
    },
    'abampere': {
        'name': _('abampere'),
        'symbol': 'abA',
    },
    'abcoulomb': {
        'name': _('abcoulomb'),
        'symbol': 'abC',
    },
    'aberdeen': {
        'name': _('Aberdeen'),
        'symbol': 'aberdeen',
    },
    'abfarad': {
        'name': _('abfarad'),
        'symbol': 'abF',
    },
    'abhenry': {
        'name': _('abhenry'),
        'symbol': 'henry',
    },
    'abohm': {
        'name': _('abohm'),
        'symbol': 'abohm',
    },
    'absiemens': {
        'name': _('absiemens'),
        'symbol': 'absiemens',
    },
    'abvolt': {
        'name': _('abvolt'),
        'symbol': 'abV',
    },
    'acre': {
        'name': _('acre'),
        'symbol': 'acre',
    },
    'acre_foot': {
        'name': _('acre-foot'),
        'symbol': 'acre-foot',
    },
    'ampere': {
        'name': _('ampere'),
        'symbol': 'A',
    },
    'milliampere': {
        'name': _('milliampere'),
        'symbol': 'mA',
    },
    'ampere_turn': {
        'name': _('ampere-turn'),
        'symbol': 'At',
    },
    'angstrom': {
        'name': _('angstrom'),
        'symbol': 'Å',
    },
    'angstrom_star': {
        'name': _('angstrom star'),
        'symbol': 'Å*',
    },
    'apothecary_dram': {
        'name': _('apothecary dram'),
        'symbol': 'dr',
    },
    'apothecary_ounce': {
        'name': _('apothecary ounce'),
        'symbol': 'oz',
    },
    'apothecary_pound': {
        'name': _('apothecary pound'),
        'symbol': 'apothecary pound',
    },
    'arcminute': {
        'name': _('minute of arc'),
        'symbol': 'arcmin',
    },
    'arcsecond': {
        'name': _('second of arc'),
        'symbol': 'arcsec',
    },
    'are': {
        'name': _('are'),
        'symbol': 'are',
    },
    'astronomical_unit': {
        'name': _('astronomical unit'),
        'symbol': 'AU',
    },
    'atmosphere_liter': {
        'name': _('atmosphere liter'),
        'symbol': 'atmosphere liter',
    },
    'atomic_mass_constant': {
        'name': _('atomic mass constant'),
        'symbol': 'Da',
    },
    'atomic_unit_of_current': {
        'name': _('atomic unit of current'),
        'symbol': '',
    },
    'atomic_unit_of_electric_field': {
        'name': _('atomic unit of electric field'),
        'symbol': '',
    },
    'atomic_unit_of_force': {
        'name': _('atomic unit of force'),
        'symbol': 'atomic unit of force',
    },
    'atomic_unit_of_intensity': {
        'name': _('atomic unit of intensity'),
        'symbol': 'atomic unit of intensity',
    },
    'atomic_unit_of_temperature': {
        'name': _('atomic unit of temperature'),
        'symbol': 'atomic unit of temperature',
    },
    'atomic_unit_of_time': {
        'name': _('atomic unit of time'),
        'symbol': 'atomic unit of time',
    },
    'avogadro_constant': {
        'name': _('Avogadro constant'),
        'symbol': 'L',
    },
    'avogadro_number': {
        'name': _('Avogadro number'),
        'symbol': '',
    },
    'bag': {
        'name': _('bag'),
        'symbol': 'bag'
    },
    'bar': {
        'name': _('bar'),
        'symbol': 'bar'
    },
    'barn': {
        'name': _('barn'),
        'symbol': 'b'
    },
    'barrel': {
        'name': _('barrel'),
        'symbol': 'barrel'
    },
    'barye': {
        'name': _('barrye'),
        'symbol': 'Ba'
    },
    'baud': {
        'name': _('baud'),
        'symbol': 'Bd'
    },
    'becquerel': {
        'name': _('becquerel'),
        'symbol': 'Bq',
    },
    'beer_barrel': {
        'name': _('beer barrel'),
        'symbol': 'beer barrel'
    },
    'bel': {
        'name': _('bel'),
        'symbol': 'B'
    },
    'biot': {
        'name': _('biot'),
        'symbol': 'biot'
    },
    'biot_turn': {
        'name': _('biot turn'),
        'symbol': 'biot-turn'
    },
    'bit': {
        'name': _('bit'),
        'symbol': 'bit'
    },
    'bits_per_pixel': {
        'name': _('bits per pixel'),
        'symbol': 'bpp'
    },
    'board_foot': {
        'name': _('board foot'),
        'symbol': 'board-foot'
    },
    'bohr': {
        'name': _('bohr'),
        'symbol': 'bohr'
    },
    'bohr_magneton': {
        'name': _('bohr magneton'),
        'symbol': 'μB'
    },
    'boiler_horsepower': {
        'name': _('boiler horsepower'),
        'symbol': 'boiler hp'
    },
    'boltzmann_constant': {
        'name': _('boltzmann constant'),
        'symbol': 'kB'
    },
    'british_thermal_unit': {
        'name': _('british thermal unit'),
        'symbol': 'Btu'
    },
    'buckingham': {
        'name': _('buckingham'),
        'symbol': 'B'
    },
    'bushel': {
        'name': _('bushel'),
        'symbol': 'bsh'
    },
    'byte': {
        'name': _('byte'),
        'symbol': 'byte'
    },
    'cables_length': {
        'name': _('cable length'),
        'symbol': 'cable'
    },
    'calorie': {
        'name': _('calorie'),
        'symbol': 'calorie'
    },
    'candela': {
        'name': _('candela'),
        'symbol': 'cd'
    },
    'carat': {
        'name': _('carat'),
        'symbol': 'ct'
    },
    'centimeter_H2O': {
        'name': _('centimeter H20'),
        'symbol': 'cm H2O'
    },
    'centimeter_Hg': {
        'name': _('centimeter Hg'),
        'symbol': 'cm Hg'
    },
    'century': {
        'name': _('century'),
        'symbol': 'c.'
    },
    'chain': {
        'name': _('chain'),
        'symbol': 'chain'
    },
    'cicero': {
        'name': _('cicero'),
        'symbol': 'cicero'
    },
    'circular_mil': {
        'name': _('circular mil'),
        'symbol': 'mils'
    },
    'classical_electron_radius': {
        'name': _('classical electron radius'),
        'symbol': 're'
    },
    'clausius': {
        'name': _('clausius'),
        'symbol': 'clausius'
    },
    'common_year': {
        'name': _('common year'),
        'symbol': ''
    },
    'conductance_quantum': {
        'name': _('conductance quantum'),
        'symbol': 'G0'
    },
    'conventional_ampere_90': {
        'name': _('conventional ampere'),
        'symbol': 'A90'
    },
    'conventional_coulomb_90': {
        'name': _('conventional coulomb'),
        'symbol': 'C90'
    },
    'conventional_farad_90': {
        'name': _('conventional farad'),
        'symbol': 'F90'
    },
    'conventional_henry_90': {
        'name': _('conventional henry'),
        'symbol': 'H90',
    },
    'conventional_josephson_constant': {
        'name': _('conventional josephson constant'),
        'symbol': 'KJ'
    },
    'conventional_ohm_90': {
        'name': _('conventional ohm'),
        'symbol': 'Ω90'
    },
    'conventional_volt_90': {
        'name': _('conventional volt'),
        'symbol': 'V90'
    },
    'conventional_von_klitzing_constant': {
        'name': _('conventional Von Klitzing constant'),
        'symbol': 'RK',
    },
    'conventional_watt_90': {
        'name': _('conventional Watt'),
        'symbol': 'W90',
    },
    'coulomb': {
        'name': _('coulomb'),
        'symbol': 'C'
    },
    'coulomb_constant': {
        'name': _('coulomb constant'),
        'symbol': 'K'
    },
    'count': {
        'name': _('count'),
        'symbol': ''
    },
    'counts_per_second': {
        'name': _('counts per second'),
        'symbol': 'counts/s'
    },
    'css_pixel': {
        'name': _('CSS pixel'),
        'symbol': 'px'
    },
    'cubic_centimeter': {
        'name': _('cubic centimeter'),
        'symbol': 'cm³'
    },
    'cubic_millimmeter': {
        'name': _('cubic milliimeter'),
        'symbol': 'mm³'
    },
    'cubic_foot': {
        'name': _('cubic foot'),
        'symbol': 'cubic foot'
    },
    'cubic_inch': {
        'name': _('cubic inch'),
        'symbol': 'cubic inch'
    },
    'cubic_yard': {
        'name': _('cubic yard'),
        'symbol': 'cubic yard'
    },
    'cup': {
        'name': _('cup'),
        'symbol': 'cup'
    },
    'curie': {
        'name': _('curie'),
        'symbol': 'Ci'
    },
    'dalton': {
        'name': _('dalton'),
        'symbol': 'Da'
    },
    'darcy': {
        'name': _('darcy'),
        'symbol': 'D'
    },
    'day': {
        'name': _('day'),
        'symbol': pgettext_lazy('unit', 'day')
    },
    'debye': {
        'name': _('debye'),
        'symbol': 'D'
    },
    'decade': {
        'name': _('decade'),
        'symbol': pgettext_lazy('symbol', 'decade')
    },
    'decibel': {
        'name': _('decibel'),
        'symbol': 'dB'
    },
    'degree': {
        'name': _('degree'),
        'symbol': '°'
    },
    'degree_Celsius': {
        'name': _('degree Celsius'),
        'symbol': '°C'
    },
    'degree_Fahrenheit': {
        'name': _('degree Fahrenheit'),
        'symbol': '°F'
    },
    'degree_Rankine': {
        'name': _('degree Rankine'),
        'symbol': '°R'
    },
    'degree_Reaumur': {
        'name': _('degree Reaumur'),
        'symbol': '°Re'
    },
    'denier': {
        'name': _('denier'),
        'symbol': 'D'
    },
    'didot': {
        'name': _('didot'),
        'symbol': 'didot'
    },
    'dirac_constant': {
        'name': _('Dirac constant'),
        'symbol': 'ħ'
    },
    'dram': {
        'name': _('dram'),
        'symbol': 'dr'
    },
    'dry_barrel': {
        'name': _('dry barrel'),
        'symbol': 'dry barrel'
    },
    'dry_gallon': {
        'name': _('dry gallon'),
        'symbol': 'dry gallon'
    },
    'dry_pint': {
        'name': _('dry pint'),
        'symbol': 'dry pint'
    },
    'dry_quart': {
        'name': _('dry quart'),
        'symbol': 'dry quart'
    },
    'dtex': {
        'name': _('decitex'),
        'symbol': 'dtex'
    },
    'dyne': {
        'name': _('dyne'),
        'symbol': 'dyn'
    },
    'electrical_horsepower': {
        'name': _('eletcrical horsepower'),
        'symbol': 'HP'
    },
    'electron_g_factor': {
        'name': _('electron G-factor'),
        'symbol': 'electron G-factor'
    },
    'electron_mass': {
        'name': _('electron mass'),
        'symbol': 'me'
    },
    'electron_volt': {
        'name': _('electronvolt'),
        'symbol': 'eV'
    },
    'elementary_charge': {
        'name': _('elementary charge'),
        'symbol': 'e'
    },
    'entropy_unit': {
        'name': _('entropy unit'),
        'symbol': 'S'
    },
    'enzyme_unit': {
        'name': _('enzyme unit'),
        'symbol': 'U'
    },
    'eon': {
        'name': _('eon'),
        'symbol': 'eon'
    },
    'erg': {
        'name': _('erg'),
        'symbol': 'erg'
    },
    'farad': {
        'name': _('farad'),
        'symbol': 'F'
    },
    'faraday': {
        'name': _('faraday'),
        'symbol': 'faraday',
    },
    'faraday_constant': {
        'name': _('faraday constant'),
        'symbol': 'F',
    },
    'fathom': {
        'name': _('fathom'),
        'symbol': 'fathom',
    },
    'fermi': {
        'name': _('fermi'),
        'symbol': 'fm',
    },
    'fifteen_degree_calorie': {
        'name': _('fifteen degree calorie'),
        'symbol': 'fifteen degree calorie',
    },
    'fifth': {
        'name': _('fifth'),
        'symbol': 'fifth',
    },
    'fine_structure_constant': {
        'name': _('fine structure constant'),
        'symbol': 'α',
    },
    'first_radiation_constant': {
        'name': _('first radiation constant'),
        'symbol': 'c1',
    },
    'fluid_dram': {
        'name': _('fluid dram'),
        'symbol': 'fluid dr'
    },
    'fluid_ounce': {
        'name': _('fluid ounce'),
        'symbol': 'fluid ounce'
    },
    'foot': {
        'name': _('foot'),
        'symbol': 'foot'
    },
    'foot_H2O': {
        'name': _('foot H2O'),
        'symbol': 'foot H2O'
    },
    'foot_per_second': {
        'name': _('foot per second'),
        'symbol': 'foot per second'
    },
    'foot_pound': {
        'name': _('foot pound'),
        'symbol': 'foot pound'
    },
    'force_gram': {
        'name': _('gram-force'),
        'symbol': 'gram-force',
        'obsolete': True,
    },
    'force_kilogram': {
        'name': _('kilogram-force'),
        'symbol': 'kg-force',
        'obsolete': True,
    },
    'force_long_ton': {
        'name': _('long ton-force'),
        'symbol': 'long ton-force',
        'obsolete': True},
    'force_metric_ton': {
        'name': _('tonne-force'),
        'symbol': 'tf',
        'obsolete': True},
    'force_ounce': {
        'name': _('ounce-force'),
        'symbol': 'ounce force',
        'obsolete': True},
    'force_pound': {
        'name': _('pound-force'),
        'symbol': 'pound force',
        'obsolete': True},
    'force_ton': {
        'name': _('tonne-force'),
        'symbol': 'tf',
        'obsolete': True},
    'fortnight': {
        'name': _('fortnight'),
        'symbol': 'fortnight'
    },
    'franklin': {
        'name': _('franklin'),
        'symbol': 'Fr'
    },
    'furlong': {
        'name': _('furlong'),
        'symbol': 'furlong'
    },
    'galileo': {
        'name': _('galileo'),
        'symbol': 'Gal'
    },
    'gallon': {
        'name': _('gallon'),
        'symbol': 'gal'
    },
    'gamma': {
        'name': _('gamma'),
        'symbol': 'γ'
    },
    'gamma_mass': {
        'name': _('gamma'),
        'symbol': 'γ'
    },
    'gauss': {
        'name': _('gauss'),
        'symbol': 'G'
    },
    'gilbert': {
        'name': _('gilbert'),
        'symbol': 'Gb'
    },
    'gill': {
        'name': _('gill'),
        'symbol': 'gill',
    },
    'grade': {
        'name': _('grade'),
        'symbol': 'gr',
    },
    'grain': {
        'name': _('grain'),
        'symbol': 'grain',
    },
    'gram': {
        'name': _('gram'),
        'symbol': 'g',
    },
    'gray': {
        'name': _('gray'),
        'symbol': 'Gy',
    },
    'gregorian_year': {
        'name': _('gregorian year'),
        'symbol': pgettext_lazy('unit', 'gregorian year'),
    },
    'hand': {
        'name': _('hand'),
        'symbol': 'hand',
    },
    'hartree': {
        'name': _('hartree'),
        'symbol': 'Eh',
    },
    'hectare': {
        'name': _('hectare'),
        'symbol': 'ha',
    },
    'henry': {
        'name': _('henry'),
        'symbol': 'henry',
    },
    'hertz': {
        'name': _('hertz'),
        'symbol': 'Hz',
    },
    'hogshead': {
        'name': _('hogshead'),
        'symbol': 'hhd',
    },
    'horsepower': {
        'name': _('horsepower'),
        'symbol': 'HP',
    },
    'hour': {
        'name': _('hour'),
        'symbol': 'h',
    },
    'hundredweight': {
        'name': _('hundredweight'),
        'symbol': 'cwt',
    },
    'impedance_of_free_space': {
        'name': _('impedance of free space'),
        'symbol': 'Z0',
    },
    'imperial_barrel': {
        'name': _('imperial barrel'),
        'symbol': 'imperial barrel',
    },
    'imperial_bushel': {
        'name': _('imperial bushel'),
        'symbol': 'bsh',
    },
    'imperial_cup': {
        'name': _('imperial cup'),
        'symbol': 'imperial cup',
    },
    'imperial_fluid_drachm': {
        'name': _('imperial fluid drachm'),
        'symbol': 'ʒ',
    },
    'imperial_fluid_ounce': {
        'name': _('imperial fluid ounce'),
        'symbol': 'fl oz',
    },
    'imperial_fluid_scruple': {
        'name': _('imperial fluid scruple'),
        'symbol': '℈',
    },
    'imperial_gallon': {
        'name': _('imperial gallon'),
        'symbol': 'imperial gallon',
    },
    'imperial_gill': {
        'name': _('imperial gill'),
        'symbol': 'imperial gill',
    },
    'imperial_minim': {
        'name': _('imperial minim'),
        'symbol': 'min',
    },
    'imperial_peck': {
        'name': _('imperial peck'),
        'symbol': 'peck'
    },
    'imperial_pint': {
        'name': _('imperial pint'),
        'symbol': 'imperial pint'
    },
    'imperial_quart': {
        'name': _('imperial quart'),
        'symbol': 'imperial quart'
    },
    'inch': {
        'name': _('inch'),
        'symbol': 'inch'
    },
    'inch_H2O_39F': {
        'name': _('inch H20 39F'),
        'symbol': 'inch H20 39F'
    },
    'inch_H2O_60F': {
        'name': _('inch H20 60F'),
        'symbol': 'inch H20 60F'
    },
    'inch_Hg': {
        'name': _('inch Hg'),
        'symbol': 'inch Hg'
    },
    'inch_Hg_60F': {
        'name': _('inch Hg 60F'),
        'symbol': 'inch Hg 60F'
    },
    'international_british_thermal_unit': {
        'name': _('international Btu'),
        'symbol': 'international Btu'
    },
    'international_calorie': {
        'name': _('international calorie'),
        'symbol': 'international calorie'
    },
    'josephson_constant': {
        'name': _('Josephson constant'),
        'symbol': 'KJ'
    },
    'joule': {
        'name': _('Joule'),
        'symbol': 'J'
    },
    'jute': {
        'name': _('jute'),
        'symbol': 'jute'
    },
    'katal': {
        'name': _('katal'),
        'symbol': 'kat'
    },
    'kelvin': {
        'name': _('kelvin'),
        'symbol': 'K'
    },
    'kilometer_per_hour': {
        'name': _('kilometer per hour'),
        'symbol': 'km/hour'
    },
    'kilometer_per_second': {
        'name': _('kilometer per second'),
        'symbol': 'km/second'
    },
    'kip': {
        'name': _('kip'),
        'symbol': 'kip'
    },
    'kip_per_square_inch': {
        'name': _('kip per square inch'),
        'symbol': 'kip / sq inch'
    },
    'knot': {
        'name': _('knot'),
        'symbol': 'kt'
    },
    'lambda': {
        'name': _('lambda'),
        'symbol': 'λ'
    },
    'lambert': {
        'name': _('lambert'),
        'symbol': 'L'
    },
    'langley': {
        'name': _('langley'),
        'symbol': 'Ly'
    },
    'lattice_spacing_of_Si': {
        'name': _('lattice constant'),
        'symbol': 'Ly'
    },
    'league': {
        'name': _('league'),
        'symbol': 'league',
        'obsolete': True},
    'leap_year': {
        'name': _('leap year'),
        'symbol': _('year')
    },
    'light_year': {
        'name': _('light year'),
        'symbol': _('ly')
    },
    'link': {
        'name': _('link'),
        'symbol': _('li')
    },
    'liter': {
        'name': _('liter'),
        'symbol': _('L')
    },
    'ln10': {
        'name': _('log 10'),
        'symbol': _('ln10')
    },
    'long_hundredweight': {
        'name': _('long hundredweight'),
        'symbol': 'long hundredweight',
        'obsolete': True},
    'long_ton': {
        'name': _('long ton'),
        'symbol': 'long ton',
        'obsolete': True},
    'lumen': {
        'name': _('lumen'),
        'symbol': 'lm'
    },
    'lux': {
        'name': _('lux'),
        'symbol': 'lx'
    },
    'magnetic_flux_quantum': {
        'name': _('magnetic flux quantum'),
        'symbol': 'Φ'
    },
    'maxwell': {
        'name': _('maxwell'),
        'symbol': 'Mx'
    },
    'mean_international_ampere': {
        'name': _('mean international ampere'),
        'symbol': 'mean international A'
    },
    'mean_international_ohm': {
        'name': _('mean international Ohm'),
        'symbol': 'mean international Ω'
    },
    'mean_international_volt': {
        'name': _('mean international Volt'),
        'symbol': 'mean international V'
    },
    'mercury': {
        'name': _('density of mercury'),
        'symbol': ''
    },
    'mercury_60F': {
        'name': _('density of mercury 60F'),
        'symbol': ''
    },
    'meter': {
        'name': _('meter'),
        'symbol': 'm'
    },
    'metric_horsepower': {
        'name': _('metric horsepower'),
        'symbol': 'hp'
    },
    'metric_ton': {
        'name': _('metric ton'),
        'symbol': 't'
    },
    'micron': {
        'name': _('micron'),
        'symbol': 'μm'
    },
    'mil': {
        'name': _('mil'),
        'symbol': 'ml'
    },
    'mile': {
        'name': _('mile'),
        'symbol': 'mile'
    },
    'mile_per_hour': {
        'name': _('mile per hour'),
        'symbol': 'mph'
    },
    'millennium': {
        'name': _('millenium'),
        'symbol': _('millenium')
    },
    'milliarcsecond': {
        'name': _('milliarcsecond'),
        'symbol': 'mas'
    },
    'millimeter_Hg': {
        'name': _('Hg millimeter'),
        'symbol': 'mmHg'
    },
    'minim': {
        'name': _('minim'),
        'symbol': 'min'
    },
    'minute': {
        'name': _('minute'),
        'symbol': 'min'
    },
    'molar': {
        'name': _('molar'),
        'symbol': 'M'
    },
    'molar_gas_constant': {
        'name': _('molar gas constant'),
        'symbol': 'R'
    },
    'mole': {
        'name': _('mole'),
        'symbol': 'mol'
    },
    'month': {
        'name': _('month'),
        'symbol': 'm'
    },
    'nautical_mile': {
        'name': _('nautical mile'),
        'symbol': _('nm')
    },
    'neper': {
        'name': _('neper'),
        'symbol': _('Np')
    },
    'neutron_mass': {
        'name': _('neutron mass'),
        'symbol': 'mN'
    },
    'newton': {
        'name': _('newton'),
        'symbol': 'N'
    },
    'newtonian_constant_of_gravitation': {
        'name': _('newtonian constant of gravitation'),
        'symbol': 'N'
    },
    'nit': {
        'name': _('nit'),
        'symbol': 'cd/m²'
    },
    'nuclear_magneton': {
        'name': _('nuclear magneton'),
        'symbol': 'μN'
    },
    'number_english': {
        'name': _('number english'),
        'symbol': 'm/g'
    },
    'number_meter': {
        'name': _('number meter'),
        'symbol': 'm/g'
    },
    'oersted': {
        'name': _('oersted'),
        'symbol': 'Oe'
    },
    'ohm': {
        'name': _('ohm'),
        'symbol': 'Ω'
    },
    'oil_barrel': {
        'name': _('oil barrel'),
        'symbol': 'oil barrel'
    },
    'ounce': {
        'name': _('ounce'),
        'symbol': 'oz t'
    },
    'parsec': {
        'name': _('parsec'),
        'symbol': 'pc'
    },
    'particle': {
        'name': _('particle'),
        'symbol': 'particle'
    },
    'pascal': {
        'name': _('pascal'),
        'symbol': 'Pa'
    },
    'peak_sun_hour': {
        'name': _('peak sun hour'),
        'symbol': 'peak sun hour'
    },
    'peck': {
        'name': _('peck'),
        'symbol': 'peck'
    },
    'pennyweight': {
        'name': _('pennyweight'),
        'symbol': 'dwt'
    },
    'pi': {
        'name': 'pi',
        'symbol': 'π'
    },
    'pica': {
        'name': _('pica'),
        'symbol': 'pica'
    },
    'pint': {
        'name': _('pint'),
        'symbol': 'pt'
    },
    'pixel': {
        'name': _('pixel'),
        'symbol': 'px'
    },
    'pixels_per_centimeter': {
        'name': _('pixel per centimeter'),
        'symbol': 'ppcm'
    },
    'pixels_per_inch': {
        'name': _('pixel per inch'),
        'symbol': 'ppi'
    },
    'planck_constant': {
        'name': _('Planck constant'),
        'symbol': 'h'
    },
    'planck_current': {
        'name': _('Planck current'),
        'symbol': _('qP')
    },
    'planck_length': {
        'name': _('Planck current'),
        'symbol': _('lP')
    },
    'planck_mass': {
        'name': _('Planck mass'),
        'symbol': _('mP')
    },
    'planck_temperature': {
        'name': _('Planck temperature'),
        'symbol': _('TP')
    },
    'planck_time': {
        'name': _('Planck time'),
        'symbol': _('tP')
    },
    'point': {
        'name': _('point'),
        'symbol': 'pt'
    },
    'poise': {
        'name': _('poise'),
        'symbol': 'P'
    },
    'pound': {
        'name': _('pound'),
        'symbol': 'lb'
    },
    'pound_force_per_square_inch': {
        'name': _('pound force per square inch'),
        'symbol': 'lb-force/inch²'
    },
    'poundal': {
        'name': _('poundal'),
        'symbol': 'pdl'
    },
    'proton_mass': {
        'name': _('proton mass'),
        'symbol': 'mP'
    },
    'quadrillion_Btu': {
        'name': _('quadrillion Btu'),
        'symbol': 'quadrillion Btu'
    },
    'quart': {
        'name': _('quart'),
        'symbol': 'qt'
    },
    'quarter': {
        'name': _('quarter'),
        'symbol': 'qr'
    },
    'radian': {
        'name': _('radian'),
        'symbol': 'rad'
    },
    'rads': {
        'name': _('rads'),
        'symbol': 'rads'
    },
    'reciprocal_centimeter': {
        'name': _('reciprocal centimeter'),
        'symbol': 'cm⁻¹'
    },
    'refrigeration_ton': {
        'name': _('refrigeration ton'),
        'symbol': 'TR'
    },
    'rem': {
        'name': _('roentgen equivalent man'),
        'symbol': 'rem'
    },
    'revolutions_per_minute': {
        'name': _('revolutions per minute'),
        'symbol': _('rpm')
    },
    'revolutions_per_second': {
        'name': _('revolutions per second'),
        'symbol': _('rps')
    },
    'reyn': {
        'name': _('reyn'),
        'symbol': _('reyn')
    },
    'rhe': {
        'name': _('reciprocal poise'),
        'symbol': 'P⁻¹'
    },
    'rod': {
        'name': _('rod'),
        'symbol': 'rod'
    },
    'roentgen': {
        'name': _('roentgen'),
        'symbol': 'R'
    },
    'rutherford': {
        'name': _('rutherford'),
        'symbol': 'Rd'
    },
    'rydberg': {
        'name': _('rydberg'),
        'symbol': 'Ry'
    },
    'rydberg_constant': {
        'name': _('rydberg constant'),
        'symbol': 'R∞'
    },
    'scaled_point': {
        'name': _('scaled point'),
        'symbol': 'Pt'
    },
    'scruple': {
        'name': _('scruple'),
        'symbol': '℈'
    },
    'second': {
        'name': _('second'),
        'symbol': 's'
    },
    'millisecond': {
        'name': _('millisecond'),
        'symbol': 'ms'
    },
    'microsecond': {
        'name': _('microsecond'),
        'symbol': 'μs'
    },
    'second_radiation_constant': {
        'name': _('second radiation constant'),
        'symbol': 'c2'
    },
    'shake': {
        'name': _('shake'),
        'symbol': 'shake'
    },
    'shot': {
        'name': _('shot'),
        'symbol': 'shot'
    },
    'sidereal_day': {
        'name': _('sidereal day'),
        'symbol': pgettext_lazy('unit', 'sidereal day')
    },
    'sidereal_month': {
        'name': _('sidereal month'),
        'symbol': pgettext_lazy('unit', 'sidereal month')
    },
    'sidereal_year': {
        'name': _('sidereal year'),
        'symbol': pgettext_lazy('unit', 'sidereal year')
    },
    'siemens': {
        'name': _('siemens'),
        'symbol': 'S'
    },
    'sievert': {
        'name': _('sievert'),
        'symbol': 'Sv'
    },
    'slinch': {
        'name': _('slinch'),
        'symbol': 'slinch'
    },
    'slug': {
        'name': _('slug'),
        'symbol': 'slug'
    },
    'speed_of_light': {
        'name': _('speed of light'),
        'symbol': 'c'
    },
    'square_degree': {
        'name': _('square degree'),
        'symbol': '(°)²'
    },
    'square_foot': {
        'name': _('square foot'),
        'symbol': 'sq ft'
    },
    'square_inch': {
        'name': _('square inch'),
        'symbol': 'sq in'
    },
    'square_league': {
        'name': _('square league'),
        'symbol': 'sq league'
    },
    'square_mile': {
        'name': _('square mile'),
        'symbol': 'sq mile'
    },
    'square_rod': {
        'name': _('square rod'),
        'symbol': 'sq rod'
    },
    'square_survey_mile': {
        'name': _('square survey mile'),
        'symbol': 'sq survey mile'
    },
    'square_yard': {
        'name': _('square yard'),
        'symbol': 'sq yd'
    },
    'standard_atmosphere': {
        'name': _('standard atmosphere'),
        'symbol': 'atm'
    },
    'standard_gravity': {
        'name': _('standard gravity'),
        'symbol': 'g0'
    },
    'standard_liter_per_minute': {
        'name': _('standard liter per minute'),
        'symbol': 'SLM'
    },
    'statampere': {
        'name': 'statampere',
        'symbol': 'statA'
    },
    'statfarad': {
        'name': 'statfarad',
        'symbol': 'statF'
    },
    'stathenry': {
        'name': 'stathenry',
        'symbol': 'stathenry'
    },
    'statmho': {
        'name': 'statmho',
        'symbol': 'statmho'
    },
    'statohm': {
        'name': 'statohm',
        'symbol': 'statohm'
    },
    'stattesla': {
        'name': 'stattesla',
        'symbol': 'stattesla'
    },
    'statvolt': {
        'name': 'statvolt',
        'symbol': 'statvolt'
    },
    'statweber': {
        'name': 'statweber',
        'symbol': 'statweber'
    },
    'stefan_boltzmann_constant': {
        'name': _('Stefan-Boltzmann constant'),
        'symbol': 'σ'
    },
    'steradian': {
        'name': _('steradian'),
        'symbol': 'sr'
    },
    'stere': {
        'name': _('stere'),
        'symbol': 'st'
    },
    'stilb': {
        'name': _('stilb'),
        'symbol': 'sb'
    },
    'stokes': {
        'name': _('stokes'),
        'symbol': 'St'
    },
    'stone': {
        'name': _('stone'),
        'symbol': 'st.'
    },
    'survey_foot': {
        'name': _('survey foot'),
        'symbol': 'survey ft'
    },
    'survey_mile': {
        'name': _('survey mile'),
        'symbol': 'survey mile'
    },
    'svedberg': {
        'name': 'svedberg',
        'symbol': 'S'
    },
    'synodic_month': {
        'name': _('synodic month'),
        'symbol': _('synodic month')
    },
    'tablespoon': {
        'name': _('tablespoon'),
        'symbol': _('tablespoon')
    },
    'tansec': {
        'name': 'tansec',
        'symbol': 'tansec'
    },
    'teaspoon': {
        'name': _('teaspoon'),
        'symbol': _('teaspoon')
    },
    'technical_atmosphere': {
        'name': _('technical atmosphere'),
        'symbol': 'at'
    },
    'tesla': {
        'name': 'tesla',
        'symbol': 'T'
    },
    'tex': {
        'name': 'tex',
        'symbol': 'tex'
    },
    'tex_cicero': {
        'name': 'tex cicero',
        'symbol': 'tex cicero'
    },
    'tex_didot': {
        'name': 'tex didot',
        'symbol': 'tex didot'
    },
    'tex_pica': {
        'name': 'tex pica',
        'symbol': 'tex pica'
    },
    'tex_point': {
        'name': 'tex point',
        'symbol': 'tex point'
    },
    'therm': {
        'name': _('therm'),
        'symbol': 'Thm'
    },
    'thermochemical_british_thermal_unit': {
        'name': _('thermochemical Btu'),
        'symbol': 'thermochemical Btu'
    },
    'thomson_cross_section': {
        'name': _('Thomson cross section'),
        'symbol': 'Thomson cross section'
    },
    'thou': {
        'name': 'thou',
        'symbol': 'thou'
    },
    'ton': {
        'name': _('ton'),
        'symbol': 't'
    },
    'ton_TNT': {
        'name': _('ton of TNT'),
        'symbol': 'ton of TNT'
    },
    'tonne_of_oil_equivalent': {
        'name': _('ton of oil equivalent'),
        'symbol': _('toe')
    },
    'torr': {
        'name': 'torr',
        'symbol': 'torr'
    },
    'tropical_month': {
        'name': _('tropical month'),
        'symbol': _('tropical month')
    },
    'tropical_year': {
        'name': _('tropical year'),
        'symbol': _('tropical year')
    },
    'troy_ounce': {
        'name': _('troy ounce'),
        'symbol': _('troy ounce')
    },
    'troy_pound': {
        'name': _('troy pound'),
        'symbol': _('troy pound')
    },
    'turn': {
        'name': _('turn'),
        'symbol': _('turn')
    },
    'unified_atomic_mass_unit': {
        'name': _('unified atomic mass unit'),
        'symbol': _('unified atomic mass unit')
    },
    'unit_pole': {
        'name': _('unit pole'),
        'symbol': 'pole'
    },
    'vacuum_permeability': {
        'name': _('vacuum permeability'),
        'symbol': 'μ0'
    },
    'vacuum_permittivity': {
        'name': _('vacuum permittivity'),
        'symbol': 'ε0'
    },
    'volt': {
        'name': 'Volt',
        'symbol': 'V'
    },
    'volt_ampere': {
        'name': 'Volt Ampere',
        'symbol': 'VA'
    },
    'von_klitzing_constant': {
        'name': _('Von Klitzing constant'),
        'symbol': 'RK'
    },
    'water': {
        'name': _('Water volumic mass'),
        'symbol': 'g/m³'
    },
    'water_39F': {
        'name': _('Water volumic mass at 39F'),
        'symbol': 'g/m³'
    },
    'water_60F': {
        'name': _('Water volumic mass at 60F'),
        'symbol': 'g/m³'
    },
    'watt': {
        'name': 'Watt',
        'symbol': 'W'
    },
    'watt_hour': {
        'name': 'Watt',
        'symbol': 'Wh'
    },
    'weber': {
        'name': 'weber',
        'symbol': 'weber'
    },
    'week': {
        'name': 'week',
        'symbol': 'w'
    },
    'wien_frequency_displacement_law_constant': {
        'name': _('wien frequency displacement law constant'),
        'symbol': 'vpeak'
    },
    'wien_u': {
        'name': _('wien u'),
        'symbol': ''
    },
    'wien_wavelength_displacement_law_constant': {
        'name': _('wien displacement law constant'),
        'symbol': 'λpeak'
    },
    'wien_x': {
        'name': _('wien x'),
        'symbol': ''
    },
    'x_unit_Cu': {
        'name': _('Cu X-ray wavelength'),
        'symbol': ''
    },
    'x_unit_Mo': {
        'name': _('Mo X-ray wavelength'),
        'symbol': ''
    },
    'yard': {
        'name': _('yard'),
        'symbol': 'yd'
    },
    'year': {
        'name': _('year'),
        'symbol': _('year')
    },
    'zeta': {
        'name': _('zeta'),
        'symbol': ''
    },
    'decibelmilliwatt': {
        'name': _('decibelmilliwatt'),
        'symbol': 'dBmW'
    },
    'decibelmicrowatt': {
        'name': _('decibelmicrowatt'),
        'symbol': 'dBμW'
    },
    'eulers_number': {
        'name': _('Euler‘s number'),
        'symbol': 'e'
    },
    'octave': {
        'name': _('octave'),
        'symbol': ''
    },
    'sound_pressure_level': {
        'name': _('sound pressure level'),
        'symbol': ''
    },
    'meter_per_second': {
        'name': _('meter per second'),
        'symbol': 'm.s⁻¹'
    },
    'meter_per_square_second': {
        'name': _('meter per square second'),
        'symbol': 'm.s⁻²'
    },
    'square_meter': {
        'name': _('square meter'),
        'symbol': 'm²'
    },
    'cubic_meter': {
        'name': _('square meter'),
        'symbol': 'm³'
    },
    'milligram': {
        'name': _('milligram'),
        'symbol': 'mg'
    },
    'kilogram': {
        'name': _('kilogram'),
        'symbol': 'kg'
    },
    'kilometer': {
        'name': _('kilometer'),
        'symbol': 'km'
    },
    'centimeter': {
        'name': _('centimeter'),
        'symbol': 'cm'
    },
    'millimeter': {
        'name': _('millimeter'),
        'symbol': 'mm'
    },
    'centimeter_per_second': {
        'name': _('centimeter per second'),
        'symbol': 'cm.s⁻¹'
    },
}



"""
Rates services
"""
from collections import Iterator

from datetime import date
from django.conf import settings


class RateNotAvailableError(Exception):
    """
    Rate could not be found
    """
    pass


class RatesNotAvailableError(Exception):
    """
    Rates service is not available
    """
    pass


class RateService:
    """
    Rate service
    """

    def serializer(self, base_currency, currency, date, value):
        """
        Serialize rate from service
        """
        return {
            'base_currency': base_currency,
            'currency': currency,
            'date': date,
            'value': value
        }

    def available_currencies(self) -> Iterator:
        """
        Return list of avalaible currencies
        """
        raise NotImplementedError

    def fetch_rates(self,
                    base_currency: str = settings.BASE_CURRENCY,
                    currency: str = None,
                    date_obj: date = date.today(),
                    to_obj: date = None) -> []:
        """
        Get rates for a base currency at a given date
        :param base_currency: Base currency
        :param date_obj: Date of value
        :param to_obj: Optional range of values
        :return: List of dicts [{'base_currency': base currency, 'currency': currency, 'date': date of value, 'value'}]
        """
        raise NotImplementedError



"""
Units models
"""

import logging
from datetime import date

import pint.systems
from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.utils.translation import ugettext as _
from geocurrency.converters.models import BaseConverter, ConverterResult, \
    ConverterResultDetail, ConverterResultError, ConverterLoadError, \
    CalculationResultError, CalculationResultDetail, CalculationResult
from sympy import sympify, SympifyError

from . import UNIT_EXTENDED_DEFINITION, DIMENSIONS, UNIT_SYSTEM_BASE_AND_DERIVED_UNITS, \
    ADDITIONAL_BASE_UNITS, PREFIX_SYMBOL
from .exceptions import *
from .settings import ADDITIONAL_UNITS, PREFIXED_UNITS_DISPLAY


class Quantity:
    """
    Quantity class
    """
    system = None
    unit = None
    value = 0
    date_obj = None

    def __init__(self, system: str, unit: str, value: float, date_obj: date = None):
        """
        Initialize quantity on unit system
        """
        self.system = system
        self.unit = unit
        self.value = value
        self.date_obj = date_obj

    def __repr__(self):
        """
        Look beautiful
        """
        return f'{self.value} {self.unit} ({self.system})'


class Unit:
    """
    Unit mock for hinting
    """
    pass


class UnitSystem:
    """
    Pint UnitRegistry wrapper
    """
    ureg = None
    system_name = None
    system = None
    _additional_units = set()

    def __init__(self, system_name: str = 'SI', fmt_locale: str = 'en', user: User = None,
                 key: str = None):
        """
        Initialize UnitSystem from name and user / key information for loading custom units
        """
        found = False
        for available_system in UnitSystem.available_systems():
            if system_name.lower() == available_system.lower():
                system_name = available_system
                found = True
        if not found:
            raise UnitSystemNotFound("Invalid unit system")
        self.system_name = system_name
        try:
            additional_units_settings = settings.GEOCURRENCY_ADDITIONAL_UNITS
        except AttributeError:
            additional_units_settings = ADDITIONAL_UNITS
        try:
            self.ureg = pint.UnitRegistry(system=system_name, fmt_locale=fmt_locale)
            self.system = getattr(self.ureg.sys, system_name)
            self._load_additional_units(units=ADDITIONAL_BASE_UNITS)
            self._load_additional_units(units=additional_units_settings)
            if user:
                self._load_custom_units(user=user, key=key)
            self._rebuild_cache()
        except (FileNotFoundError, AttributeError):
            raise UnitSystemNotFound("Invalid unit system")

    def _rebuild_cache(self):
        """
        Rebuild registry cache
        It should be in the define method of the registry
        """
        self.ureg._build_cache()

    def _load_additional_units(self, units: dict, redefine: bool = False) -> bool:
        """
        Load additional base units in registry
        """
        available_units = self.available_unit_names()
        if not self.system_name in units:
            logging.warning(f"error loading additional units for {self.system_name}")
            return False
        added_units = []
        for key, items in units[self.system_name].items():
            if key not in available_units:
                self.ureg.define(f"{key} = {items['relation']} = {items['symbol']}")
                added_units.append(key)
            elif redefine:
                self.ureg.redefine(f"{key} = {items['relation']} = {items['symbol']}")
        self._additional_units = self._additional_units | set(added_units)
        return True

    def _load_custom_units(self, user: User, key: str = None, redefine: bool = False) -> bool:
        """
        Load custom units in registry
        """
        if user and user.is_authenticated:
            if user.is_superuser:
                qs = CustomUnit.objects.all()
            else:
                qs = CustomUnit.objects.filter(user=user)
        if key:
            qs = qs.filter(key=key)
        qs = qs.filter(unit_system=self.system_name)
        available_units = self.available_unit_names()
        added_units = []
        for cu in qs:
            props = [cu.code, cu.relation]
            if cu.symbol:
                props.append(cu.symbol)
            if cu.alias:
                props.append(cu.alias)
            definition = " = ".join(props)
            if cu.code not in available_units:
                self.ureg.define(definition)
                added_units.append(cu.code)
            elif redefine:
                self.ureg.redefine(definition)
            else:
                logging.error(f"{cu.code} already defined in registry")
        self._additional_units = self._additional_units | set(added_units)
        return True

    def _test_additional_units(self, units: dict) -> bool:
        """
        Load and check dimensionality of ADDITIONAL_BASE_UNITS values
        """
        if self.system_name not in units:
            return False
        for key in units[self.system_name].keys():
            try:
                test = self.unit(key).dimensionality and True
            except pint.errors.UndefinedUnitError:
                return False
        return True

    def add_definition(self, code, relation, symbol, alias):
        """
        Add a new unit definition to a UnitSystem, and rebuild cache
        :param code: code of the unit
        :param relation: relation to other units (e.g.: 3 kg/m)
        """
        self.ureg.define(f"{code} = {relation} = {symbol} = {alias}")
        self._rebuild_cache()

    @classmethod
    def available_systems(cls) -> [str]:
        """
        List of available Unit Systems
        :return: Array of string
        """
        ureg = pint.UnitRegistry(system='SI')
        return dir(ureg.sys)

    @classmethod
    def is_valid(cls, system: str) -> bool:
        """
        Check validity of the UnitSystem
        :param name: name of the unit system
        """
        us = cls()
        return system in us.available_systems()

    def current_system(self) -> pint.UnitRegistry:
        """
        Return current pint.UnitRegistry
        """
        return self.ureg

    def unit(self, unit_name):
        """
        Create a Object in the UnitSystem
        :param unit_name: name of the unit in the unit system
        """
        return Unit(unit_system=self, code=unit_name)

    def available_unit_names(self) -> [str]:
        """
        List of available units for a given Unit system
        :return: Array of names of Unit systems
        """
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        prefixed_units = []
        for key, prefixes in prefixed_units_display.items():
            for prefix in prefixes:
                prefixed_units.append(prefix + key)
        return sorted(prefixed_units + dir(getattr(self.ureg.sys, self.system_name))
                      + list(self._additional_units))

    def unit_dimensionality(self, unit: str) -> str:
        """
        User friendly representation of the dimension
        :param unit: name of the unit to display
        :return: Human readable dimension
        """
        return Unit.dimensionality_string(unit_system=self.system, unit_str=unit)

    def available_dimensions(self, ordering: str = 'name') -> {}:
        """
        Return available dimensions for the UnitSystem
        :param ordering: sort result by attribute
        """
        if ordering not in ['name', 'code', 'dimension']:
            ordering = 'name'
        return sorted([Dimension(unit_system=self, code=dim) for dim in DIMENSIONS.keys()],
                      key=lambda x: getattr(x, ordering, ''))

    @property
    def _ureg_dimensions(self):
        """
        return dimensions with units
        """
        dimensions = []
        for dim in self.ureg._dimensions:
            try:
                if not self.ureg.get_compatible_units(dim):
                    continue
                dimensions.append(dim)
            except KeyError:
                continue
        return dimensions

    def _get_dimension_dimensionality(self, dimension: str) -> {}:
        """
        Return the dimensionality of a dimension based on the first compatible unit
        """
        for dim in self.ureg.get_compatible_units(dimension):
            return self.ureg.get_base_units(dim)[1]

    def _generate_dimension_delta_dictionnary(self) -> {}:
        """
        Generate the dict to put in DIMENSIONS
        """
        output = {}
        for dim in self._ureg_dimensions:
            if not dim in DIMENSIONS:
                output[dim] = {
                    'name': f'_({dim})',
                    'dimension': str(self._get_dimension_dimensionality(dim)),
                    'symbol': ''
                }
        return output

    def units_per_dimension(self, dimensions: [str]) -> {}:
        """
        Return units grouped by dimension
        :param dimensions: restrict list of dimensions
        """
        output = {}
        registry_dimensions = dimensions or DIMENSIONS.keys()
        for dim in registry_dimensions:
            dimension = Dimension(unit_system=self, code=dim)
            try:
                if units := self.ureg.get_compatible_units(dim):
                    output[dim] = units
            except KeyError:
                continue
        return output

    def units_per_dimensionality(self) -> {}:
        """
        List of units per dimension
        :return: dict of dimensions, with lists of unit strings
        """
        units_array = self.available_unit_names()
        output = {}
        for unit_str in units_array:
            dimension = Unit.dimensionality_string(self, unit_str)
            try:
                output[dimension].append(unit_str)
            except KeyError:
                output[dimension] = [unit_str]
        return output

    @property
    def dimensionalities(self) -> [str]:
        """
        List of dimensions available in the Unit system
        :return: list of dimensions for Unit system
        """
        return set([Unit.dimensionality_string(self, unit_str) for unit_str in dir(self.system)])


class Dimension:
    """
    Dimenion of a Unit
    """
    unit_system = None
    code = None
    name = None
    dimension = None

    def __init__(self, unit_system: UnitSystem, code: str):
        """
        Initialize a Dimension in a UnitSystem
        """
        try:
            dimension = DIMENSIONS[code]
            self.unit_system = unit_system
            self.code = code
            self.name = dimension['name']
            self.dimension = dimension['dimension']
        except (ValueError, KeyError) as e:
            logging.warning(str(e))
            self.code = None
        if not self.code:
            raise DimensionNotFound

    def __repr__(self):
        """
        Look beautiful
        """
        return self.code

    def units(self, user=None, key=None) -> [Unit]:
        """
        List of units for this dimension
        :param user: optional user for custom units
        :param key: optional key for custom units
        """
        if self.code == '[compounded]':
            return self._compounded_units
        if self.code == '[custom]':
            return self._custom_units(user=user, key=key)
        unit_list = []
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        try:
            unit_list.append(
                self.unit_system.unit(
                    UNIT_SYSTEM_BASE_AND_DERIVED_UNITS[self.unit_system.system_name][self.code]
                )
            )
        except (KeyError, UnitNotFound):
            logging.warning(f"unable to find base unit for"
                            f"unit system {self.unit_system.system_name}"
                            f" and dimension {self.code}")
        unit_list.extend(
            [
                Unit(unit_system=self.unit_system, pint_unit=unit)
                for unit in self.unit_system.ureg.get_compatible_units(self.code)
            ])
        unit_names = [str(u) for u in unit_list]
        for unit, prefixes in prefixed_units_display.items():
            if unit in unit_names:
                for prefix in prefixes:
                    unit_list.append(self.unit_system.unit(unit_name=prefix + unit))
        return set(sorted(unit_list, key=lambda x: x.name))

    @property
    def _compounded_units(self):
        """
        List units that do not belong to a dimension
        """
        available_units = self.unit_system.available_unit_names()
        dimensioned_units = []
        for dimension_code in [d for d in DIMENSIONS.keys() if
                               d != '[compounded]' and d != '[custom]']:
            dimension = Dimension(unit_system=self.unit_system, code=dimension_code)
            dimensioned_units.extend([u.code for u in dimension.units()])
        return [self.unit_system.unit(au) for au in set(available_units) - set(dimensioned_units)]

    def _custom_units(self, user: User, key: str = None) -> [Unit]:
        """
        Return list of custom units
        :param user: User owning the units
        :param key: optional unit key
        """
        if user and user.is_authenticated:
            if user.is_superuser:
                custom_units = CustomUnit.objects.all()
            else:
                custom_units = CustomUnit.objects.filter(user=user)
            if key:
                custom_units = custom_units.filter(key=key)
            return [self.unit_system.unit(cu.code) for cu in custom_units]
        else:
            return []

    @property
    def base_unit(self):
        """
        Base unit for this dimension in this Unit System
        """
        try:
            return UNIT_SYSTEM_BASE_AND_DERIVED_UNITS[self.unit_system.system_name][self.code]
        except KeyError:
            logging.warning(
                f'dimension {self.dimension} is not part of unit system {self.unit_system.system_name}')
            return None


class Unit:
    """
    Pint Unit wrapper
    """
    unit_system = None
    code = None
    unit = None

    def __init__(self, unit_system: UnitSystem, code: str = '', pint_unit: pint.Unit = None):
        """
        Initialize a Unit in a UnitSystem
        :param unit_system: UnitSystem instance
        :param code: code of the pint.Unit
        """
        self.unit_system = unit_system
        if pint_unit and isinstance(pint_unit, pint.Unit):
            self.code = str(pint_unit)
            self.unit = pint_unit
        elif code:
            self.code = code
            try:
                self.unit = getattr(unit_system.system, code)
            except pint.errors.UndefinedUnitError:
                raise UnitNotFound("invalid unit for system")
        else:
            raise UnitNotFound("invalid unit for system")

    def __repr__(self):
        return self.code

    @classmethod
    def is_valid(cls, name: str) -> bool:
        """
        Check the validity of a unit in a UnitSystem
        """
        try:
            us_si = UnitSystem(system_name='SI')
        except UnitSystemNotFound:
            return False
        try:
            return us_si.unit(unit_name=name) and True
        except pint.errors.UndefinedUnitError:
            return False

    @property
    def name(self) -> str:
        """
        Return name of the unit from table of units
        """
        return self.unit_name(self.code)

    @property
    def symbol(self) -> str:
        """
        Return symbol for Unit
        """
        return self.unit_symbol(self.code)

    @property
    def dimensions(self) -> [Dimension]:
        """
        Return Dimensions of Unit
        """
        dimensions = [Dimension(unit_system=self.unit_system, code=code) for code in
                      DIMENSIONS.keys()
                      if DIMENSIONS[code]['dimension'] == str(self.dimensionality)]
        return dimensions or '[compounded]'

    def base_unit(unit_str: str) -> (str, str):
        """
        Get base unit in case the unit is a prefixed unit
        :param unit_str: name of unit to check
        :return: base unit name, prefix
        """
        prefix = ''
        base_str = unit_str
        try:
            prefixed_units_display = settings.GEOCURRENCY_PREFIXED_UNITS_DISPLAY
        except AttributeError:
            prefixed_units_display = PREFIXED_UNITS_DISPLAY
        for base, prefixes in prefixed_units_display.items():
            for _prefix in prefixes:
                if unit_str == _prefix + base:
                    prefix = _prefix
                    base_str = base
        return base_str, prefix

    @staticmethod
    def unit_name(unit_str: str) -> str:
        """
        Get translated name from unit string
        :param unit_str: Name of unit
        """
        base_str, prefix = Unit.base_unit(unit_str=unit_str)
        try:
            ext_unit = UNIT_EXTENDED_DEFINITION.get(base_str)
            return prefix + str(ext_unit['name'])
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {base_str}')
            return unit_str

    @staticmethod
    def unit_symbol(unit_str: str) -> str:
        """
        Static function to get symbol from unit string
        :param unit_str: Name of unit
        """
        base_str, prefix = Unit.base_unit(unit_str=unit_str)
        try:
            prefix_symbol = PREFIX_SYMBOL[prefix]
            ext_unit = UNIT_EXTENDED_DEFINITION.get(base_str)
            return prefix_symbol + ext_unit['symbol']
        except (KeyError, TypeError) as e:
            logging.error(f'No UNIT_EXTENDED_DEFINITION for unit {base_str}')
            return ''

    @staticmethod
    def dimensionality_string(unit_system: UnitSystem, unit_str: str) -> str:
        """
        Converts pint dimensionality string to human readable string
        :param unit_system: UnitSystem
        :param unit_str: Unit name
        :return: str
        """
        ds = str(getattr(unit_system.ureg, unit_str).dimensionality).replace('[', '').replace(']',
                                                                                              '')
        ds = ds.replace(' ** ', '^')
        ds = ds.split()
        return ' '.join([_(d) for d in ds])

    @property
    def dimensionality(self):
        """
        Return dimensionality of a unit in Pint universe
        """
        try:
            return self.unit_system.ureg.get_base_units(self.code)[1]
        except KeyError:
            return ''

    @staticmethod
    def translated_name(unit_system: UnitSystem, unit_str: str) -> str:
        """
        Translated name of the unit
        """
        try:
            return '{}'.format(unit_system.ureg[unit_str])
        except KeyError:
            return unit_str

    @property
    def readable_dimension(self):
        """
        Wrapper around Unit.dimensionality_string
        """
        return Unit.dimensionality_string(unit_system=self.unit_system, unit_str=self.code)


class UnitConverter(BaseConverter):
    """
    Conversion between units
    """
    base_system = None
    base_unit = None
    user = None
    key = None

    def __init__(self, base_system: str, base_unit: str, user: User = None, key: key = None,
                 id: str = None):
        """
        Initialize the converter. It converts a payload into a destination unit
        """
        try:
            super().__init__(id=id)
            self.base_system = base_system
            self.base_unit = base_unit
            self.user = user
            self.key = key
            self.system = UnitSystem(system_name=base_system, user=user, key=key)
            self.unit = Unit(unit_system=self.system, code=base_unit)
            self.compatible_units = [str(u) for u in self.unit.unit.compatible_units()]
        except (UnitSystemNotFound, UnitNotFound):
            raise UnitConverterInitError

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super().add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        system = str
        unit = str
        value = float
        date_obj ('YYYY-MM-DD')
        """
        from .serializers import QuantitySerializer
        errors = []
        for line in data:
            serializer = QuantitySerializer(data=line)
            if serializer.is_valid():
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(cls, id: str, user: User = None, key: str = None) -> BaseConverter:
        """
        Load converter from ID
        """
        try:
            uc = super().load(id)
            uc.system = UnitSystem(system_name=uc.base_system, user=user, key=key)
            uc.unit = Unit(unit_system=uc.system, code=uc.base_unit)
            return uc
        except (UnitSystemNotFound, UnitNotFound, KeyError) as e:
            raise ConverterLoadError

    def save(self):
        """
        Save the converter to cache
        """
        system = self.system
        unit = self.unit
        self.system = None
        self.unit = None
        super().save()
        self.system = system
        self.unit = unit

    def convert(self) -> ConverterResult:
        """
        Converts data to base unit in base system
        """

        result = ConverterResult(id=self.id, target=self.base_unit)
        Q_ = self.system.ureg.Quantity
        for quantity in self.data:
            if quantity.unit not in self.compatible_units:
                error = ConverterResultError(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    error=_('Incompatible units')
                )
                result.errors.append(error)
                continue
            try:
                pint_quantity = Q_(quantity.value, quantity.unit)
                out = pint_quantity.to(self.base_unit)
                result.increment_sum(out.magnitude)
                detail = ConverterResultDetail(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    conversion_rate=0,
                    converted_value=out.magnitude
                )
                result.detail.append(detail)
            except pint.UndefinedUnitError:
                error = ConverterResultError(
                    unit=quantity.unit,
                    original_value=quantity.value,
                    date=quantity.date_obj,
                    error=_('Undefined unit in the registry')
                )
                result.errors.append(error)
        self.end_batch(result.end_batch())
        return result


class UnitConversionPayload:
    """
    Unit conversion payload
    """
    data = None
    base_system = ''
    base_unit = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, base_system: UnitSystem, base_unit: Unit, data=None, key: str = None,
                 batch_id: str = None, eob: bool = False):
        """
        Initialize conversion payload
        """
        self.data = data
        self.base_system = base_system
        self.base_unit = base_unit
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class CustomUnit(models.Model):
    """
    Additional unit for a user
    """
    AVAILABLE_SYSTEMS = (
        ('Planck', 'Planck'),
        ('SI', 'SI'),
        ('US', 'US'),
        ('atomic', 'atomic'),
        ('cgs', 'CGS'),
        ('imperial', 'imperial'),
        ('mks', 'mks'),
    )
    user = models.ForeignKey(User, related_name='units', on_delete=models.PROTECT)
    key = models.CharField(max_length=255, default=None, db_index=True, null=True, blank=True)
    unit_system = models.CharField(max_length=20, choices=AVAILABLE_SYSTEMS)
    code = models.SlugField()
    name = models.CharField("Human readable name", max_length=255)
    relation = models.CharField("Relation to an existing unit", max_length=255)
    symbol = models.CharField("Symbol", max_length=20, blank=True, null=True)
    alias = models.CharField("Alias", max_length=20, null=True, blank=True)

    class Meta:
        """
        Meta
        """
        unique_together = ('user', 'key', 'code')

    def save(self, *args, **kwargs):
        """
        Save custom unit to database
        """
        us = UnitSystem(system_name=self.unit_system)
        if self.code in us.available_unit_names():
            raise UnitDuplicateError
        try:
            us.add_definition(code=self.code, relation=self.relation, symbol=self.symbol,
                              alias=self.alias)
        except ValueError as e:
            raise UnitValueError(str(e)) from e
        try:
            us.unit(self.code).unit.dimensionality
        except pint.errors.UndefinedUnitError:
            raise UnitDimensionError
        return super(CustomUnit, self).save(*args, **kwargs)


class Operand:
    """
    Operand in a formula
    """
    name = None
    value = None
    unit = None

    def __init__(self, name=None, value=None, unit=None):
        """
        Initialize Operand
        :param name: name of the operand that appears in the formula
        :param value: float magnitude
        :param unit: units
        """
        self.name = name
        self.value = value
        self.unit = unit

    def validate(self):
        """
        Validate operand
        """
        if not self.name or self.value is None or self.unit is None:
            return False
        return True


class ComputationError(Exception):
    pass


class Expression:
    """
    Expression with operands
    """
    expression = None
    operands = None

    def __init__(self, expression: str, operands: [Operand]):
        """
        Initialize Expression
        :param expression: string expression with placeholders
        :param operands: List of Operand corresponding to placeholders
        """
        self.expression = expression
        self.operands = operands

    def validate(self, unit_system: UnitSystem) -> (bool, str):
        """
        Validate syntax and homogeneity
        :param unit_system: UnitSystem
        """
        Q_ = unit_system.ureg.Quantity
        if not self.expression:
            return False, "missing expression"
        try:
            sympify(self.expression.format(**{v.name: v.value for v in self.operands}))
        except SympifyError as e:
            return False, "Improper expression"
        for var in self.operands:
            if not var.validate():
                return False, "invalid operand"
        kwargs = {v.name: f"{v.value} {v.unit}" for v in self.operands}
        try:
            Q_(self.expression.format(**kwargs))
        except KeyError:
            return False, "Missing operand"
        except pint.errors.DimensionalityError:
            return False, "Incoherent dimensions"
        return True, ''

    def evaluate(self, unit_system: UnitSystem) -> pint.Quantity:
        """
        Validate formula
        """
        Q_ = unit_system.ureg.Quantity
        is_valid, error = self.validate(unit_system=unit_system)
        if is_valid:
            kwargs = {v.name: f"{v.value} {v.unit}" for v in self.operands}
            return Q_(self.expression.format(**kwargs))
        else:
            raise ComputationError(f"Invalid formula: {error}")


class CalculationPayload:
    """
    Calculation payload: caculate expressions
    """
    data = None
    unit_system = ''
    user = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, unit_system: UnitSystem, key: str = None, data: [] = None,
                 batch_id: str = None, eob: bool = False):
        """
        Initialize payload
        """
        self.data = data
        self.unit_system = unit_system
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class ExpressionCalculator(BaseConverter):
    """
    Conversion between units
    """
    unit_system = None

    def __init__(self, unit_system: str, user: User = None, key: str = '', id: str = None):
        """
        Initiate ExpressionCalculator
        :param unit_system: unit system name
        :param user: User
        :param key: key of user
        :param id: ID of the batch
        """
        try:
            super().__init__(id=id)
            self.unit_system = unit_system
            self.user = user
            self.key = key
            self.system = UnitSystem(system_name=unit_system, user=self.user, key=self.key)
        except (UnitSystemNotFound) as e:
            raise ExpressionCalculatorInitError from e

    def add_data(self, data: []) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super().add_data(data)
        return errors

    def check_data(self, data):
        """
        Validates that the data contains
        - system (str)
        - unit (str)
        - value (float)
        - date (YYYY-MM-DD)
        """
        from .serializers import ExpressionSerializer
        errors = []
        for line in data:
            serializer = ExpressionSerializer(data=line)
            if serializer.is_valid(unit_system=self.system):
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    @classmethod
    def load(cls, id: str, user: User = None, key: str = None) -> BaseConverter:
        """
        Load converter from batch
        :param id: ID of the batch
        """
        try:
            uc = super().load(id)
            uc.system = UnitSystem(system_name=uc.unit_system, user=user, key=key)
            return uc
        except (UnitSystemNotFound, KeyError) as e:
            raise ConverterLoadError

    def save(self):
        """
        Save converter to caching system
        """
        system = self.system
        self.system = None
        super(ExpressionCalculator, self).save()
        self.system = system

    def convert(self) -> CalculationResult:
        """
        Converts data to base unit in base system
        """
        result = CalculationResult(id=self.id)
        for expression in self.data:
            valid, exp_error = expression.validate(unit_system=self.system)
            if not valid:
                error = CalculationResultError(
                    expression=expression.unit,
                    operands=expression.operands,
                    date=date.today(),
                    error=exp_error
                )
                result.errors.append(error)
                continue
            try:
                out = expression.evaluate(unit_system=self.system)
            except ComputationError as e:
                error = CalculationResultError(
                    expression=expression.unit,
                    operands=expression.operands,
                    date=date.today(),
                    error=str(e)
                )
                result.errors.append(error)
                continue
            detail = CalculationResultDetail(
                expression=expression.expression,
                operands=expression.operands,
                magnitude=out.magnitude,
                units=out.units
            )
            result.detail.append(detail)
        self.end_batch(result.end_batch())
        return result



"""
List of serializers for Converters and Calculators
"""

from rest_framework import serializers

from .models import Batch, ConverterResultError, ConverterResultDetail, ConverterResult, \
    CalculationResult, CalculationResultError, CalculationResultDetail


class CalculationResultDetailSerializer(serializers.Serializer):
    """
    Serializer for the CalculationResultDetail class
    """
    expression = serializers.CharField()
    operands = serializers.JSONField()
    magnitude = serializers.FloatField()
    units = serializers.CharField()

    def create(self, validated_data):
        """
        Create a CalculationResultDetail object
        :param validated_data: cleaned data
        """
        return CalculationResultDetail(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResultDetail object
        :param instance: CalculationResultDetail object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression', instance.expression)
        instance.operands = validated_data.get('operands', instance.variables)
        instance.date = validated_data.get('date', instance.date)
        return instance


class CalculationResultErrorSerializer(serializers.Serializer):
    """
    Serialiwer for the CalculationResultError class
    """
    expression = serializers.CharField()
    variables = serializers.JSONField()
    date = serializers.DateField()
    error = serializers.CharField()

    def create(self, validated_data):
        """
        Create a CalculationResultError object
        :param validated_data: cleaned data
        """
        return CalculationResultError(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResultError object
        :param instance: CalculationResultError object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression', instance.expression)
        instance.operands = validated_data.get('operands', instance.variables)
        instance.date = validated_data.get('date', instance.date)
        instance.error = validated_data.get('error', instance.error)
        return instance


class CalculationResultSerializer(serializers.Serializer):
    id = serializers.UUIDField()
    detail = CalculationResultDetailSerializer(many=True)
    status = serializers.CharField()
    errors = CalculationResultErrorSerializer(many=True)

    def create(self, validated_data):
        """
        Create a CalculationResult object
        :param validated_data: cleaned data
        """
        return CalculationResult(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a CalculationResult object
        :param instance: CalculationResult object
        :param validated_data: cleaned data
        """
        instance.id = validated_data.get('id', instance.id)
        instance.detail = validated_data.get('detail', instance.detail)
        instance.status = validated_data.get('status', instance.status)
        instance.errors = validated_data.get('errors', instance.errors)
        return instance


class ConverterResultDetailSerializer(serializers.Serializer):
    """
    Serialize a ConverterResultDetail class
    """
    unit = serializers.CharField()
    original_value = serializers.FloatField()
    date = serializers.DateField()
    conversion_rate = serializers.FloatField()
    converted_value = serializers.FloatField()

    def create(self, validated_data):
        """
        Create a ConverterResultDetail object
        :param validated_data: cleaned data
        """
        return ConverterResultDetail(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResultDetail object
        :param instance: ConverterResultDetail object
        :param validated_data: cleaned data
        """
        instance.unit = validated_data.get('unit', instance.unit)
        instance.original_value = validated_data.get('original_value', instance.original_value)
        instance.date = validated_data.get('date', instance.date)
        return instance


class ConverterResultErrorSerializer(serializers.Serializer):
    """
    Serializer for ConverterResultError
    """
    unit = serializers.CharField()
    original_value = serializers.FloatField()
    date = serializers.DateField()
    error = serializers.CharField()

    def create(self, validated_data):
        """
        Create a ConverterResultError object
        :param validated_data: cleaned data
        """
        return ConverterResultError(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResultError object
        :param instance: ConverterResultError object
        :param validated_data: cleaned data
        """
        instance.unit = validated_data.get('unit', instance.unit)
        instance.original_value = validated_data.get('original_value', instance.original_value)
        instance.date = validated_data.get('date', instance.date)
        instance.error = validated_data.get('error', instance.error)
        return instance


class ConverterResultSerializer(serializers.Serializer):
    """
    Serializer for a ConverterResult
    """
    id = serializers.UUIDField()
    target = serializers.CharField()
    detail = ConverterResultDetailSerializer(many=True)
    sum = serializers.FloatField()
    status = serializers.CharField()
    errors = ConverterResultErrorSerializer(many=True)

    def create(self, validated_data):
        """
        Create a ConverterResult object
        :param validated_data: cleaned data
        """
        return ConverterResult(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a ConverterResult object
        :param instance: ConverterResult object
        :param validated_data: cleaned data
        """
        instance.id = validated_data.get('id', instance.id)
        instance.target = validated_data.get('target', instance.target)
        instance.detail = validated_data.get('detail', instance.detail)
        instance.sum = validated_data.get('sum', instance.sum)
        instance.status = validated_data.get('status', instance.status)
        instance.errors = validated_data.get('errors', instance.errors)
        return instance


class BatchSerializer(serializers.Serializer):
    """
    Serializer for Batch
    """
    id = serializers.CharField()
    status = serializers.CharField()

    def create(self, validated_data):
        """
        Create a Batch object
        :param validated_data: cleaned data
        """
        return Batch(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Batch
        :param instance: Batch object
        :param validated_data: cleaned data
        """
        instance.status = validated_data.get('status', instance.status)
        return instance



"""
Filter classes for Currencies module
"""
from django_filters import rest_framework as filters

from .models import Currency


class CurrencyFilter(filters.FilterSet):
    """
    Filter on Currency
    """
    name = filters.CharFilter(field_name='name', lookup_expr='icontains')
    exponent = filters.NumberFilter(field_name='exponent')
    country = filters.CharFilter(field_name='countries__name', lookup_expr='icontains')
    exact_country = filters.CharFilter(field_name='countries__name', lookup_expr='iexact')
    capital = filters.CharFilter(field_name='countries__capital', lookup_expr='icontains')

    class Meta:
        model = Currency
        fields = [
            'code',
            'name',
            'exponent',
            'country',
            'exact_country',
            'capital'
        ]



"""
Models for Rates module
"""
from datetime import date, timedelta

import networkx as nx
from django.conf import settings
from django.contrib.auth.models import User
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils.translation import ugettext_lazy as _
from geocurrency.converters.models import BaseConverter, ConverterResult, ConverterResultDetail, \
    ConverterResultError
from geocurrency.core.helpers import service

try:
    RATE_SERVICE = settings.RATE_SERVICE
except AttributeError:
    pass

from .services import RatesNotAvailableError


class NoRateFound(Exception):
    """
    Exception when no rate is found
    """
    pass


class BaseRate(models.Model):
    """
    Just an abstract for value and hinting
    """
    value = None

    class Meta:
        """
        Yes, it is abstract
        """
        abstract = True


class RateManager(models.Manager):
    """
    Manager for Rate model
    """

    @staticmethod
    def __sync_rates__(rates: [], base_currency: str):
        """
        Sync rates to the database
        :param rates: array of dict of rates from service
        :param base_currency: base currency to fetch
        """
        output = []
        for rate in rates:
            _rate, created = Rate.objects.get_or_create(
                base_currency=base_currency,
                currency=rate.get('currency'),
                value_date=rate.get('date'),
                user=None,
                key=None
            )
            _rate.value = rate.get('value')
            _rate.save()
            output.append(_rate)
        return output

    def fetch_rates(self,
                    base_currency: str,
                    currency: str = None,
                    rate_service: str = None,
                    date_obj: date = date.today(),
                    to_obj: date = None) -> []:
        """
        Get rates from a service for a base currency and stores them in the database
        :param rate_service: Service class to fetch from
        :param currency: currency to obtain rate for
        :param base_currency: base currency to get rate from
        :param date_obj: date to fetch rates at
        :param to_obj: end of range
        :return: QuerySet of Rate
        """
        service_name = rate_service or settings.RATE_SERVICE
        try:
            rates = service(service_type='rates', service_name=service_name).fetch_rates(
                base_currency=base_currency,
                currency=currency,
                date_obj=date_obj,
                to_obj=to_obj)
            if not rates:
                return False
        except RatesNotAvailableError:
            return False
        return self.__sync_rates__(rates=rates, base_currency=base_currency)

    def rate_at_date(self,
                     currency: str,
                     key: str = None,
                     base_currency: str = settings.BASE_CURRENCY,
                     date_obj: date = date.today()) -> BaseRate:
        """
        Get a rate at a given date
        :param currency: target currency
        :param base_currency: destination currency
        :param key: User key
        :param date_obj: date at which to fetch the rate
        """
        rate = self.find_rate(
            currency=currency,
            key=key,
            base_currency=base_currency,
            date_obj=date_obj)
        if rate.pk:
            return rate
        return Rate()

    @classmethod
    def currency_shortest_path(cls, currency: str, base_currency: str, key: str = None,
                               date_obj: date = date.today()) -> [str]:
        """
        Return the shortest path between 2 currencies for the given date
        :param currency: source currency code
        :param base_currency: base currency code
        :param key: Key specific to a client
        :param date_obj: Date to obtain the conversion rate for
        :return List of currency codes to go from currency to base currency
        """
        rates = Rate.objects.filter(value_date=date_obj).filter(
            models.Q(user=None) | models.Q(key=key))
        rates_couples = rates.values('currency', 'base_currency', 'value', 'key')
        graph = nx.Graph()
        for k in rates_couples:
            weight = 0.5 if k['base_currency'] == 'EUR' or k['currency'] == 'EUR' else 1
            weight *= (0.5 if k['key'] else 1)
            graph.add_edge(u_of_edge=k['currency'], v_of_edge=k['base_currency'], weight=weight)
        try:
            return nx.shortest_path(graph, currency, base_currency, weight='weight')
        except nx.exception.NetworkXNoPath as exc:
            raise NoRateFound(
                f"Rate {currency} to {base_currency} on "
                f"key {key} at date {date_obj} does not exist") \
                from exc

    def find_rate(self, currency: str,
                  rate_service: str = None,
                  key: str = None,
                  base_currency: str = settings.BASE_CURRENCY,
                  date_obj: date = date.today(),
                  use_forex: bool = False) -> BaseRate:
        """
        Find rate based on Floyd Warshall algorithm
        :param currency: source currency code
        :param base_currency: base currency code
        :param key: Key specific to a client
        :param date_obj: Date to obtain the conversion rate for
        :param rate_service: Rate service to use
        :param use_forex: use rate service to fill the gaps
        """
        if use_forex:
            if not self.fetch_rates(
                    base_currency=base_currency,
                    currency=currency,
                    rate_service=rate_service,
                    date_obj=date_obj):
                return Rate()
        try:
            rates = self.currency_shortest_path(
                currency=currency,
                base_currency=base_currency,
                key=key,
                date_obj=date_obj
            )
        except NoRateFound:
            # No relation found, try fetching rate
            return self.find_rate(
                currency=currency,
                rate_service=rate_service,
                key=key,
                base_currency=base_currency,
                date_obj=date_obj,
                use_forex=True
            )
        # Direct connection between rates
        if len(rates) == 2:
            return Rate.objects.filter(
                currency=currency,
                base_currency=base_currency,
                value_date=date_obj
            ).filter(
                models.Q(key=key) | models.Q(user__isnull=True)
            ).order_by('-key').first()
        else:
            conv_value = 1
            for i in range(len(rates) - 1):
                from_cur, to_cur = rates[i:i + 2]
                if rate := self.find_rate(
                        currency=from_cur,
                        base_currency=to_cur,
                        key=key,
                        date_obj=date_obj,
                        use_forex=False):
                    conv_value *= rate.value
                else:
                    raise NoRateFound(
                        f"rate {from_cur} -> {to_cur} for key {key} "
                        f"does not exist at date {date_obj}")
            rate = Rate.objects.create(
                key=key,
                value_date=date_obj,
                currency=currency,
                base_currency=base_currency,
                value=conv_value
            )
            return rate


class Rate(BaseRate):
    """
    Class Rate
    """
    user = models.ForeignKey(User, related_name='rates', on_delete=models.PROTECT, null=True)
    key = models.CharField(max_length=255, default=None, db_index=True, null=True)
    value_date = models.DateField()
    value = models.FloatField(default=0)
    currency = models.CharField(max_length=3, db_index=True)
    base_currency = models.CharField(max_length=3, db_index=True, default='EUR')
    objects = RateManager()

    class Meta:
        """
        Meta
        """
        ordering = ['-value_date', ]
        indexes = [
            models.Index(fields=['base_currency', 'value_date']),
            models.Index(fields=['currency', 'value_date']),
            models.Index(fields=['currency', 'base_currency', 'value_date']),
            models.Index(fields=['key', 'currency', 'base_currency', 'value_date']),
        ]
        unique_together = [['key', 'currency', 'base_currency', 'value_date']]

    @classmethod
    def convert(cls, currency: str,
                user: User = None,
                key: str = None,
                base_currency: str = 'EUR',
                date_obj: date = date.today(),
                amount: float = 0) -> ConverterResult:
        """
        Convert rate
        :param user: Django User
        :param key: key for user
        :param base_currency: destination currency
        :param currency: source currency
        :param date_obj: date of the rate
        :param amount: amount to convert
        """
        converter = RateConverter(user=user, key=key, base_currency=base_currency)
        converter.add_data(
            {
                'currency': currency,
                'amount': amount,
                'date': date_obj
            }
        )
        result = converter.convert()
        return result


@receiver(post_save, sender=Rate)
def create_reverse_rate(sender, instance, created, **kwargs):
    """
    Create the rate object to revert rate when create a rate
    """
    if created and not Rate.objects.filter(
            user=instance.user,
            key=instance.key,
            value_date=instance.value_date,
            currency=instance.base_currency,
            base_currency=instance.currency,
    ).exists() and instance.value != 0:
        Rate.objects.create(
            user=instance.user,
            key=instance.key,
            value_date=instance.value_date,
            currency=instance.base_currency,
            base_currency=instance.currency,
            value=1 / instance.value
        )


class Amount:
    """
    Amount with a currency, a value and a date
    """
    currency = None
    amount = 0
    date_obj = None

    def __init__(self, currency: str, amount: float, date_obj: date):
        """
        Initialize amount
        """
        self.currency = currency
        self.amount = amount
        self.date_obj = date_obj

    def __repr__(self):
        """
        How do I look like
        """
        return f'{self.date_obj}: {self.currency} {self.amount}'


class RateConversionPayload:
    """
    Payload for conversion of amounts
    """
    data = None
    target = ''
    key = ''
    batch_id = ''
    eob = False

    def __init__(self, target, data=None, key=None, batch_id=None, eob=False):
        """
        Representation of the payload
        """
        self.data = data
        self.target = target
        self.key = key
        self.batch_id = batch_id
        self.eob = eob


class BulkRate:
    """
    Rate that is applied to a range of dates
    """
    base_currency = settings.BASE_CURRENCY
    currency = settings.BASE_CURRENCY
    value = 0
    key = None
    from_date = None
    to_date = None

    def __init__(self, base_currency, currency, value, key, from_date, to_date):
        """
        Initialize
        :param key: key for user
        :param base_currency: destination currency
        :param currency: source currency
        """
        self.base_currency = base_currency
        self.currency = currency
        self.value = value
        self.key = key
        self.from_date = from_date
        self.to_date = to_date

    def to_rates(self, user):
        """
        Create rates in the database
        """
        if not self.to_date:
            self.to_date = date.today()
        rates = []
        for i in range((self.to_date - self.from_date).days + 1):
            rate, created = Rate.objects.get_or_create(
                user=user,
                key=self.key,
                base_currency=self.base_currency,
                currency=self.currency,
                value_date=self.from_date + timedelta(i)
            )
            rate.value = self.value
            rate.save()
            rates.append(rate)
        return rates


class RateConverter(BaseConverter):
    """
    Converter of rates
    """
    base_currency = settings.BASE_CURRENCY
    cached_currencies = {}
    user = None
    key = None

    def __init__(self, user: User, id: str = None, key: str = None,
                 base_currency: str = settings.BASE_CURRENCY):
        """
        Initialize
        :param user: Django User
        :param key: key for user
        :param base_currency: destination currency
        """
        super(RateConverter, self).__init__(id=id)
        self.base_currency = base_currency
        self.user = user
        self.key = key

    def add_data(self, data: [Amount]) -> []:
        """
        Check data and add it to the dataset
        Return list of errors
        """
        errors = super(RateConverter, self).add_data(data)
        self.cache_currencies()
        return errors

    def check_data(self, data: []) -> []:
        """
        Validates that the data contains
        - currency (str)
        - amount (float)
        - date (YYYY-MM-DD)
        """
        from .serializers import RateAmountSerializer
        errors = []
        for line in data:
            serializer = RateAmountSerializer(data=line)
            if serializer.is_valid():
                self.data.append(serializer.create(serializer.validated_data))
            else:
                errors.append(serializer.errors)
        return errors

    def cache_currencies(self):
        """
        Reads currencies in data and fetches rates, put them in memory
        """
        for line in self.data:
            self.cached_currencies[line.date_obj] = self.cached_currencies.get(line.date_obj) or {}
            rate = Rate.objects.rate_at_date(
                key=self.key,
                base_currency=self.base_currency,
                currency=line.currency,
                date_obj=line.date_obj)
            if rate.pk:
                self.cached_currencies[line.date_obj][line.currency] = rate.value

    def convert(self) -> ConverterResult:
        """
        Converts data to base currency
        """
        result = ConverterResult(id=self.id, target=self.base_currency)
        for amount in self.data:
            rate = self.cached_currencies[amount.date_obj][amount.currency]
            if rate:
                value = float(amount.amount) / rate
                result.increment_sum(value)
                detail = ConverterResultDetail(
                    unit=amount.currency,
                    original_value=amount.amount,
                    date=amount.date_obj,
                    conversion_rate=rate,
                    converted_value=value
                )
                result.detail.append(detail)
            else:
                error = ConverterResultError(
                    unit=amount.currency,
                    original_value=amount.amount,
                    date=amount.date_obj,
                    error=_('Rate could not be found')
                )
                result.errors.append(error)
        self.end_batch(result.end_batch())
        return result



"""
Serializers for Units module
"""

import json
import logging
from datetime import date, datetime

import pint
from drf_yasg.utils import swagger_serializer_method
from geocurrency.core.serializers import UserSerializer
from rest_framework import serializers
from sympy import sympify, SympifyError

from .models import Quantity, UnitConversionPayload, Dimension, \
    CustomUnit, Expression, CalculationPayload, UnitSystem, Operand, Unit


class QuantitySerializer(serializers.Serializer):
    """
    Serialize a Quantity
    """
    system = serializers.CharField()
    unit = serializers.CharField()
    value = serializers.FloatField()
    date_obj = serializers.DateField()

    @staticmethod
    def validate_system(system):
        """
        Validate unit system by name
        :param system: system name
        """
        from .models import UnitSystem
        if not UnitSystem.is_valid(system):
            raise serializers.ValidationError('Invalid system')
        return system

    @staticmethod
    def validate_unit(unit):
        """
        Validating unit requires knowledge of the system that might embark user information for
        custom units
        Let‘s say it is valid for now
        """
        return unit

    @staticmethod
    def validate_date_obj(value):
        """
        Validate date
        :param value:
        """
        if isinstance(value, date):
            return value
        try:
            datetime.strptime(value, 'YYYY-MM-DD')
        except ValueError:
            raise serializers.ValidationError('Invalid date format, use YYYY-MM-DD')
        return value

    def create(self, validated_data):
        """
        Create a Quantity object
        :param validated_data: cleaned data
        """
        return Quantity(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Quantity object
        :param instance: Quantity object
        :param validated_data: cleaned data
        """
        instance.system = validated_data.get('system', instance.system)
        instance.unit = validated_data.get('unit', instance.unit)
        instance.value = validated_data.get('value', instance.value)
        instance.date_obj = validated_data.get('date_obj', instance.date_obj)
        return instance


class UnitSerializer(serializers.Serializer):
    """
    Unit class serializer
    """
    code = serializers.CharField()
    name = serializers.CharField()
    dimensions = serializers.SerializerMethodField()

    @swagger_serializer_method(serializer_or_field=serializers.ListField)
    def get_dimensions(self, obj):
        """
        Get dimension of unit
        :param obj: Unit instance
        """
        return str(obj.dimensions)


class DimensionSerializer(serializers.Serializer):
    """
    Dimension class serializer
    """
    code = serializers.CharField()
    name = serializers.CharField()
    dimension = serializers.CharField()
    base_unit = serializers.SerializerMethodField()

    def create(self, validated_data):
        """
        Create a Dimension object from cleaned data
        :param validated_data: cleaned data
        """
        return Dimension(**validated_data)

    def update(self, instance, validated_data):
        """
        Update a Dimension object
        :param instance: Dimension object
        :param validated_data: cleaned data
        """
        instance.unit_system = validated_data.get('system', instance.unit_system)
        instance.code = validated_data.get('code', instance.code)
        instance.value = validated_data.get('name', instance.name)
        instance.dimension = validated_data.get('date_obj', instance.dimension)
        return instance

    @swagger_serializer_method(serializer_or_field=serializers.CharField)
    def get_base_unit(self, obj: Dimension) -> Unit:
        """
        Get base unit for dimension object
        :param obj: Dimension object
        """
        return obj.base_unit


class DimensionWithUnitsSerializer(DimensionSerializer):
    """
    Serialize a Dimension including associated units
    """
    units = UnitSerializer(many=True)

    @swagger_serializer_method(serializer_or_field=UnitSerializer)
    def get_units(self, obj: Dimension) -> [Unit]:
        """
        Get units for this dimension
        """
        try:
            return obj.units()
        except KeyError as e:
            logging.error(str(e))
            return []


class UnitSystemSerializer(serializers.Serializer):
    """
    Serialize a UnitSystem
    """
    system_name = serializers.CharField()


class UnitConversionPayloadSerializer(serializers.Serializer):
    """
    Serialize a UnitConversionPayload
    """
    data = QuantitySerializer(many=True, required=False)
    base_system = serializers.CharField(required=True)
    base_unit = serializers.CharField(required=True)
    batch_id = serializers.CharField(required=False)
    key = serializers.CharField(required=False)
    eob = serializers.BooleanField(default=False)

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of a converson payload
        """
        super().is_valid(raise_exception=raise_exception)
        if not self.initial_data.get('data') and (not self.initial_data.get('batch_id')
                                                  or (self.initial_data.get('batch_id')
                                                      and not self.initial_data.get('eob'))):
            self._errors = {
                'data': 'data has to be provided if batch_id is not provided'
                        'or batch_id is provided and eob is False'
            }

        if self._errors and raise_exception:
            raise serializers.ValidationError(self.errors)
        return not bool(self._errors)

    @staticmethod
    def validate_base_system(value: str):
        """
        Validate unit system
        :param value: unit system name
        """
        from geocurrency.units.models import UnitSystem
        if not UnitSystem.is_valid(value):
            raise serializers.ValidationError('Invalid unit system')
        return value

    @staticmethod
    def validate_base_unit(value: str) -> str:
        """
        Validate base unit
        :param value: name of unit
        """
        from geocurrency.units.models import Unit
        if not Unit.is_valid(value):
            raise serializers.ValidationError('Invalid unit')
        return value

    def create(self, validated_data: {}) -> UnitConversionPayload:
        """
        Create a UnitConversionPayload
        :param validated_data: cleaned data
        """
        return UnitConversionPayload(**validated_data)

    def update(self, instance, validated_data: {}) -> UnitConversionPayload:
        """
        Update a UnitConversionPayload
        :param instance: UnitConversionPayload
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.base_system = validated_data.get('base_system', instance.base_system)
        self.base_unit = validated_data.get('base_system', instance.base_unit)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance


class CustomUnitSerializer(serializers.ModelSerializer):
    """
    Serialize a custom unit
    """
    id = serializers.CharField(read_only=True)
    user = UserSerializer(read_only=True)
    unit_system = serializers.CharField(read_only=True)

    class Meta:
        """
        Meta
        """
        model = CustomUnit
        fields = ['id', 'user', 'key', 'unit_system', 'code', 'name', 'relation', 'symbol', 'alias']


class OperandSerializer(serializers.Serializer):
    """
    Serialize an Operand
    """
    name = serializers.CharField()
    value = serializers.CharField()
    unit = serializers.CharField()

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of the operand
        """
        if not self.initial_data.get('name'):
            raise serializers.ValidationError("variable Name not set")
        if 'value' not in self.initial_data:
            raise serializers.ValidationError("variable Value not set")
        if 'unit' not in self.initial_data:
            raise serializers.ValidationError("variable Unit not set")
        return super().is_valid(raise_exception=raise_exception)

    def create(self, validated_data):
        """
        Create an Operand object
        :param validated_data: cleaned date
        """
        return Operand(**validated_data)


class ExpressionListSerializer(serializers.ListSerializer):
    """
    Serialize a list of Expressions
    """

    def is_valid(self, unit_system: UnitSystem, raise_exception=False):
        """
        Check validity of expression with a specific UnitSystem
        :param unit_system: UnitSystem instance
        :param raise_exception: raise an exception instead of a list of errors
        """
        super().is_valid(raise_exception=raise_exception)
        for initial_exp in self.initial_data:
            ec = ExpressionSerializer(data=initial_exp)
            ec.is_valid(unit_system=unit_system)
            if ec.errors:
                self._errors.append(ec.errors)
        return not bool(self._errors)


class ExpressionSerializer(serializers.Serializer):
    """
    Serialize an expression
    """
    unit_system = None
    key = None
    expression = serializers.CharField(required=True)
    operands = OperandSerializer(many=True, required=True)

    class Meta:
        """
        Meta
        """
        list_serializer_class = ExpressionListSerializer

    def is_valid(self, unit_system: UnitSystem, raise_exception=False) -> bool:
        """
        Check if expression is valid
        :param unit_system: UnitSystem instance
        :param raise_exception: raise an exception instead of a list of errors
        """
        Q_ = unit_system.ureg.Quantity
        super().is_valid(raise_exception=raise_exception)
        try:
            operands = self.initial_data.get('operands')
            if not self.validate_operands(operands):
                self._errors['operands'] = "Invalid operands"
        except json.JSONDecodeError as e:
            self._errors['operands'] = f"Invalid operands json format: {e}"
        expression = self.initial_data.get('expression')
        if not expression:
            self._errors['expression'] = "Empty expression"
            return not bool(self._errors)
        # Validate syntax
        value_kwargs = {v['name']: v['value'] for v in operands}
        try:
            sympify(expression.format(**value_kwargs))
        except (SympifyError, KeyError) as e:
            self._errors['expression'] = f"Invalid operation"
        # Validate units
        units_kwargs = {v['name']: f"{v['value']} {v['unit']}" for v in operands}
        try:
            Q_(expression.format(**units_kwargs))
        except KeyError as e:
            self._errors['operands'] = "Missing operands"
        except pint.errors.DimensionalityError as e:
            self._errors['expression'] = "Incoherent dimension"
        if self._errors and raise_exception:
            raise ValidationError(self.errors)
        return not bool(self._errors)

    @staticmethod
    def validate_operands(value):
        """
        Validate Operand
        :param value: Obperand
        """
        for var in value:
            vs = OperandSerializer(data=var)
            vs.is_valid()
        return value

    def create(self, validated_data):
        """
        Create an Expression
        :param validated_data: cleaned data
        """
        operands = []
        for var in validated_data.get('operands'):
            vs = OperandSerializer(data=var)
            if vs.is_valid():
                operands.append(vs.create(vs.validated_data))
        return Expression(
            expression=validated_data.get('expression'),
            operands=operands
        )

    def update(self, instance, validated_data):
        """
        Update Expression
        :param instance: Expression object
        :param validated_data: cleaned data
        """
        instance.expression = validated_data.get('expression')
        instance.operands = validated_data.get('operands')
        return instance


class CalculationPayloadSerializer(serializers.Serializer):
    """
    Serializer for CalculationPayload
    """
    unit_system = serializers.CharField(required=True)
    data = ExpressionSerializer(many=True, required=False)
    batch_id = serializers.CharField(required=False)
    key = serializers.CharField(required=False)
    eob = serializers.BooleanField(default=False)

    def is_valid(self, raise_exception=False) -> bool:
        """
        Check validity of the payload
        """
        if not self.initial_data.get('data') and (not self.initial_data.get('batch_id')
                                                  or (self.initial_data.get('batch_id')
                                                      and not self.initial_data.get('eob'))):
            raise serializers.ValidationError(
                'data has to be provided if batch_id is not provided '
                'or batch_id is provided and eob is False'
            )
        return super().is_valid(raise_exception=raise_exception)

    @staticmethod
    def validate_unit_system(value: str):
        """
        Validate unit system
        :param value: unit system name
        """
        from geocurrency.units.models import UnitSystem
        if not UnitSystem.is_valid(value):
            raise serializers.ValidationError('Invalid unit system')
        return value

    def create(self, validated_data: {}):
        """
        Create CalculationPayload from cleaned data
        :param validated_data: cleaned data
        """
        return CalculationPayload(**validated_data)

    def update(self, instance, validated_data: {}):
        """
        Update a CalculationPayload from cleaned data
        :param instance: CalculationPayload object
        :param validated_data: cleaned data
        """
        self.data = validated_data.get('data', instance.data)
        self.unit_system = validated_data.get('unit_system', instance.unit_system)
        self.batch_id = validated_data.get('batch_id', instance.batch_id)
        self.key = validated_data.get('key', instance.key)
        self.eob = validated_data.get('eob', instance.eob)
        return instance



# Generated by Django 3.1.5 on 2021-03-18 07:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rates', '0002_auto_20201013_2244'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='rate',
            index=models.Index(fields=['base_currency', 'value_date'], name='rates_rate_base_cu_2508b5_idx'),
        ),
        migrations.AddIndex(
            model_name='rate',
            index=models.Index(fields=['currency', 'value_date'], name='rates_rate_currenc_4d48ff_idx'),
        ),
    ]



"""
Pagination class
"""
from rest_framework import pagination
from django.conf import settings
from .settings import MAX_PAGE_SIZE


class PageNumberPagination(pagination.PageNumberPagination):
    """
    Paginate with a page number
    """
    page_size_query_param = 'page_size'
    max_page_size = getattr(settings, 'GEOCURRENCY_MAX_PAGE_SIZE', MAX_PAGE_SIZE)




"""
Country API viewsets
"""

from countryinfo import CountryInfo
from django.conf import settings
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.decorators.cache import cache_page
from django.views.decorators.vary import vary_on_cookie
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from geocurrency.core.helpers import service
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND
from rest_framework.viewsets import ViewSet

from .models import Country, CountryNotFoundError
from .serializers import CountrySerializer, CountryDetailSerializer


class CountryViewset(ViewSet):
    """
    View for currency
    """
    lookup_field = 'alpha_2'

    language_header = openapi.Parameter('Accept-Language', openapi.IN_HEADER,
                                        description="language",
                                        type=openapi.TYPE_STRING)
    language = openapi.Parameter('language', openapi.IN_QUERY, description="language",
                                 type=openapi.TYPE_STRING)
    ordering = openapi.Parameter('ordering', openapi.IN_QUERY, description="ordering",
                                 type=openapi.TYPE_STRING)

    countries_response = openapi.Response('List of countries', CountrySerializer)
    country_detail_response = openapi.Response('Country detail', CountryDetailSerializer)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header, ordering],
                         responses={200: countries_response})
    def list(self, request):
        """
        List countries. this view is not paginated
        """
        countries = Country.all_countries(ordering=request.GET.get('ordering', 'name'))
        serializer = CountrySerializer(countries, many=True, context={'request': request})
        return Response(serializer.data)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @swagger_auto_schema(manual_parameters=[language, language_header],
                         responses={200: country_detail_response})
    def retrieve(self, request, alpha_2: str):
        """
        Retrieve a Country based on its alpha 2 code
        """
        try:
            country = Country(alpha_2)
            serializer = CountryDetailSerializer(country, context={'request': request})
            return Response(serializer.data, content_type="application/json")
        except CountryNotFoundError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='timezones', url_name='timezones')
    def timezones(self, request, alpha_2):
        """
        Send timezones for a specific country
        """
        try:
            c = Country(alpha_2)
            return Response(c.timezones, content_type="application/json")
        except KeyError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='currencies', url_name='currencies')
    def currencies(self, request, alpha_2):
        """
        Send timezones for a specific country
        """
        try:
            c = CountryInfo(alpha_2)
            return Response(c.currencies(), content_type="application/json")
        except KeyError:
            return Response(_("Unknown country or no info for this country"),
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='borders', url_name='borders')
    def borders(self, request, alpha_2):
        """
        Send borders for a specific country
        """
        try:
            c = CountryInfo(alpha_2)
            return Response(c.borders(), content_type="application/json")
        except KeyError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='provinces', url_name='provinces')
    def provinces(self, request, alpha_2):
        """
        Send provinces for a specific country
        """
        try:
            c = CountryInfo(alpha_2)
            return Response(c.provinces(), content_type="application/json")
        except KeyError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='languages', url_name='languages')
    def languages(self, request, alpha_2):
        """
        Send languages for a specific country
        """
        try:
            c = CountryInfo(alpha_2)
            return Response(c.languages(), content_type="application/json")
        except KeyError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    @method_decorator(cache_page(60 * 60 * 24))
    @method_decorator(vary_on_cookie)
    @action(['GET'], detail=True, url_path='colors', url_name='colors')
    def colors(self, request, alpha_2):
        """
            Get existing flag colors
        """
        try:
            c = Country(alpha_2=alpha_2)
            return Response(c.colors(), content_type="application/json")
        except CountryNotFoundError:
            return Response("Unknown country or no info for this country",
                            status=HTTP_404_NOT_FOUND)

    geocoder = openapi.Parameter('geocoder', openapi.IN_QUERY, description="Geocoder type",
                                 type=openapi.TYPE_STRING)
    geocoder_api_key = openapi.Parameter('geocoder_api_key', openapi.IN_QUERY,
                                         description="Geocoder API key",
                                         type=openapi.TYPE_STRING)
    address = openapi.Parameter('address', openapi.IN_QUERY, description="Address to look for",
                                type=openapi.TYPE_STRING)
    lat = openapi.Parameter('latitude', openapi.IN_QUERY, description="Latitude",
                            type=openapi.TYPE_STRING)
    lng = openapi.Parameter('longitude', openapi.IN_QUERY, description="Longitude",
                            type=openapi.TYPE_STRING)

    @swagger_auto_schema(method='get', responses={200: openapi.TYPE_ARRAY})
    @action(['GET'], detail=False, url_path='geocoders', url_name='geocoders')
    def geocoders(self, request):
        """
        Return a list of available geocoders. As defined in settings.GEOCODING_SERVICE_SETTINGS
        """
        return Response(settings.SERVICES.get('geocoding', {}).keys(),
                        content_type="application/json")

    @swagger_auto_schema(method='get', manual_parameters=[address, geocoder, geocoder_api_key])
    @action(['GET'], detail=False, url_path='geocode', url_name='geocoding')
    def geocode(self, request):
        """
        Find country by geocoding (giving address or POI)
        """
        geocoder = service(service_type='geocoding',
                           service_name=request.GET.get('geocoder', settings.GEOCODING_SERVICE),
                           key=request.GET.get('geocoder_api_key', settings.GEOCODER_GOOGLE_KEY))
        data = geocoder.search(address=request.GET.get('address'))
        countries = geocoder.countries(data)
        serializer = CountrySerializer(countries, many=True, context={'request': request})
        return Response(serializer.data)

    @swagger_auto_schema(method='get', manual_parameters=[lat, lng, geocoder, geocoder_api_key])
    @action(['GET'], detail=False, url_path='reverse', url_name='reverse_geocoding')
    def reverse_geocode(self, request):
        """
        Find country by reverse geocoding (giving latitude and longitude)
        """
        geocoder = service(service_type='geocoding',
                           service_name=request.GET.get('geocoder', settings.GEOCODING_SERVICE),
                           key=request.GET.get('geocoder_api_key', settings.GEOCODER_GOOGLE_KEY))
        data = geocoder.reverse(
            lat=request.GET.get('lat'),
            lng=request.GET.get('lon')
        )
        countries = geocoder.countries(data)
        serializer = CountrySerializer(countries, many=True, context={'request': request})
        return Response(serializer.data)



"""GeoCurrencies URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/3.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
import os

import geocurrency
from django.conf.urls import url, include
from django.contrib import admin
from django.urls import path
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from geocurrency.core import urls as geocurrency_urls
from rest_framework import permissions

environment = os.environ.get('GEOCURRENCY_ENV', 'dev')
contact_email = os.environ.get('GEOCURRENCY_CONTACT', 'fm@peabytes.me')
api_url = os.environ.get('GEOCURRENCY_URL', 'https://dev.geocurrency.me/')

schema_view = get_schema_view(
    openapi.Info(
        title=f"GeoCurrency {environment} API",
        default_version='v1',
        description="APIs for GeoCurrency v" + '.'.join(map(str, geocurrency.__version__[:3])),
        terms_of_service="/tos/",
        contact=openapi.Contact(email=contact_email),
        license=openapi.License(name="MIT License"),
    ),
    url=api_url,
    public=True,
    permission_classes=(permissions.AllowAny,),
)

urlpatterns = [
    path('admin/', admin.site.urls),
    path('accounts/', include('django.contrib.auth.urls')),
    url(r'^swagger(?P<format>\.json|\.yaml)$', schema_view.without_ui(cache_timeout=0),
        name='schema-json'),
    url(r'^swagger/$', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
    url(r'^redoc/$', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
    path('', include(geocurrency_urls))
]



"""
Command to fetch rates for all currencies
"""

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """
    Fetch rates command
    """
    help = 'Fetch rates for a date'

    def add_arguments(self, parser):
        """
        Add date argument to the command
        """
        parser.add_argument('date', nargs='*', type=str)

    def handle(self, *args, **options):
        """
        Handle call
        """
        from geocurrency.currencies.models import Currency
        from geocurrency.rates.models import Rate

        for currency in Currency.all_currencies():
            self.stdout.write('fetching rates for currency {}.'.format(currency.code))
            Rate.objects.fetch_rates(base_currency=currency.code)



"""
Currencies models
"""
import logging
from datetime import date
from typing import Iterator

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import models
from geocurrency.countries.models import Country
from iso4217 import Currency as Iso4217

from . import CURRENCY_SYMBOLS, DEFAULT_SYMBOL


class CurrencyNotFoundError(Exception):
    """
    Exception for not found currency
    """
    msg = 'Currency not found'


class Currency:
    """
    Mock for hinting
    """
    pass


class Currency:
    """
    Currency class, wrapper for ISO-4217
    """
    code = None
    name = None
    currency_name = None
    exponent = None
    number = 0
    value = None

    def __init__(self, code):
        """
        Initialize an iso4217.Currency instance
        """
        try:
            i = Iso4217(code)
            for a in ['code', 'name', 'currency_name', 'exponent', 'number', 'value']:
                setattr(self, a, getattr(i, a))
        except ValueError:
            raise CurrencyNotFoundError('Invalid currency code')

    @classmethod
    def is_valid(cls, cur: str) -> bool:
        """
        Checks if currency is part of iso4217
        """
        try:
            Currency(cur)
            return True
        except CurrencyNotFoundError:
            return False

    @classmethod
    def all_currencies(cls, ordering: str = 'name') -> Iterator[Currency]:
        """
        Returns a sorted list of currencies
        :param ordering: sort attribute
        """
        if ordering not in ['code', 'name', 'currency_name', 'exponent', 'number', 'value']:
            ordering = 'name'
        return sorted([Currency(c.code) for c in Iso4217], key=lambda x: getattr(x, ordering))

    @property
    def countries(self) -> Iterator[Country]:
        """
        List countries using this currency
        :return: List of Country objects
        """
        countries = []
        if cached_countries := cache.get(self.code + 'COUNTRIES'):
            return [Country(alpha_2) for alpha_2 in cached_countries]
        for country in Country.all_countries():
            try:
                if self.code in country.currencies():
                    countries.append(country)
            except KeyError:
                # Some countries listed in pycountry are not present in countryinfo
                pass
        cache.set(self.code + 'COUNTRIES', [country.alpha_2 for country in countries])
        return countries

    @classmethod
    def get_for_country(cls, alpha2: str) -> Iterator[Currency]:
        """
        Return a list of currencies for an alpha2 country code
        :params alpha2: alpha2 code of a country
        """
        try:
            country = Country(alpha2)
            return [Currency(cur) for cur in country.currencies()]
        except CurrencyNotFoundError as e:
            logging.error("Error fetching currency")
            logging.error(e)
            return []

    def get_rates(self, user: User = None, key: str = None, base_currency: str = None,
                  start_date: date = None,
                  end_date: date = None) -> Iterator:
        """
        Return a list of rates for this currency and an optional base currency between two dates
        :params base_currency: code of the base currency
        :params start_date: rates from that date included
        :params end_date: rates to that date included
        :return: List of rates
        """
        from geocurrency.rates.models import Rate
        qs = Rate.objects.filter(currency=self.code)
        if user:
            qs = qs.filter(models.Q(user=user) | models.Q(user=None))
        if key:
            qs = qs.filter(key=key)
        if base_currency:
            qs = qs.filter(base_currency=base_currency)
        if start_date:
            qs = qs.filter(value_date__gte=start_date)
        if end_date:
            qs = qs.filter(value_date__lte=end_date)
        return qs

    @property
    def symbol(self):
        """
        Returns the symbol for this currency based on __init__.CURRENCY_SYMBOLS
        """
        return CURRENCY_SYMBOLS.get(self.code, DEFAULT_SYMBOL)



"""
Permissions for Currency API
"""

from rest_framework import permissions


class CurrencyObjectPermission(permissions.BasePermission):
    """
    Permissions for Currency API
    """

    def has_object_permission(self, request, view, obj):
        """
        Read only permission
        """
        if request.method in permissions.SAFE_METHODS:
            return True
        return False



"""
Google Geocoder service
"""
import json
import logging

import requests
from django.conf import settings

from . import Geocoder

GEOCODER_GOOGLE_URL = 'https://maps.googleapis.com/maps/api'


class GoogleGeocoder(Geocoder):
    """
    Google geocoder class
    search and reverse from Google
    """
    coder_type = 'google'
    key = None

    def __new__(cls, *args, **kwargs):
        """
        Initialize
        """
        return super(Geocoder, cls).__new__(cls)

    def __init__(self, key:str = None, *args, **kwargs):
        """
        Google geocode engine
        Init: Geocoder('google', 'API key')
        """
        try:
            if not key or not settings.GEOCODER_GOOGLE_KEY:
                raise ValueError(
                    "This geocoder needs an API key, please provide a key"
                    " or set GEOCODER_GOOGLE_KEY in configuration")
        except AttributeError:
            raise ValueError(
                "This geocoder needs an API key, please provide a key "
                "or set GEOCODER_GOOGLE_KEY in configuration")
        self.key = key or settings.GEOCODER_GOOGLE_KEY

    def search(self, address: str, language: str = None, bounds=None,
               region: str = None, components: str = "") -> dict:
        """
        Google geocoding search
        Retrieves coordinates based on address
        """
        try:
            response = requests.get('{}/{}'.format(
                GEOCODER_GOOGLE_URL,
                'geocode/json'
            ), {
                'address': address,
                'key': self.key,
                'language': language
            })
            data = response.json()
            if data.get('status') != "OK":
                return {}
            return data
        except json.JSONDecodeError as e:
            logging.error("Invalid response")
            logging.error(e)
        except ValueError as e:
            logging.error("Invalid API configuration")
            logging.error(e)
        except IOError as e:
            logging.error("Invalid request")
            logging.error(e)
        return {}

    def reverse(self, lat:str, lng:str) -> dict:
        """
        Google geocoding reverse
        :param lat: latitude
        :param lng: longitude
        """
        try:
            response = requests.get('{}/{}'.format(
                GEOCODER_GOOGLE_URL,
                'geocode/json'
            ),
                {
                    'latlng': ",".join(map(str, [lat, lng])),
                    'key': self.key,
                })
            data = response.json()
            if data.get('status') != "OK":
                return {}
            return data
        except json.JSONDecodeError as e:
            logging.error("Invalid response")
            logging.error(e)
        except ValueError as e:
            logging.error("Invalid API configuration")
            logging.error(e)
        except IOError as e:
            logging.error("Invalid request")
            logging.error(e)
        return {}

    def parse_countries(self, data: dict) -> [str]:
        """
        parse response from google service
        :params data: geocoding / reverse geocoding json
        :return: array of alpha2 codes
        """
        alphas = []
        if not data:
            return alphas
        for feature in data.get('results'):
            for address_component in feature.get('address_components'):
                if 'country' in address_component.get('types'):
                    alphas.append(address_component.get('short_name'))
        return alphas



"""
Country services
"""
from timezonefinder import TimezoneFinder

from geocurrency.countries.models import Country, CountryNotFoundError

tf = TimezoneFinder(in_memory=True)


class Geocoder:
    """
    Geocoder services
    """

    def search(self, address, language=None, bounds=None, region=None, components=""):
        """
        Search an address
        :params text: address to search for
        :params language: optional, language of result
        :params bounds: optional, limit results to bounds
        :params region: optional, limit results to region
        :params components: optional, a components filter with elements separated by a pipe (|)
        :returns: Country
        """
        raise NotImplementedError("Use specific implementation")

    def reverse(self, lat, lng):
        """
        Search from GPS coordinates
        :params lat: latitude
        :params lng: longitude
        """
        raise NotImplementedError("Use specific implementation")

    def parse_countries(self, data):
        """
        Parse countries from result
        :params data: geocoding / reverse geocoding result
        :returns: Country instance
        """
        raise NotImplementedError("Use specific implementation")

    def countries(self, data):
        """
        List countries
        :params data: json response from geocoding / reverse geocoding service
        """
        countries = []
        alphas = self.parse_countries(data=data)
        for alpha in set(alphas):
            try:
                if len(alpha) == 2:
                    country = Country(alpha)
                elif len(alpha) == 3:
                    country = Country(alpha[0:2])
                else:
                    country = Country(alpha)
            except CountryNotFoundError:
                continue
            countries.append(country)
        return sorted(countries, key=lambda x: x.name)



"""
Countries Django application
"""
from django.apps import AppConfig


class CountryConfig(AppConfig):
    """
    CountryConfig class
    """
    name = 'countries'
    verbose_name = "Countries query app"

    def ready(self):
        """
        Setup config
        """
        super(CountryConfig, self).ready()



"""
Core serializers
"""
from rest_framework import serializers


class UserSerializer(serializers.BaseSerializer):
    """
    Serialize Django Auth User
    """
    username = serializers.CharField()
    email = serializers.CharField()

    def to_representation(self, instance):
        """
        Representation of the user
        """
        return {
            'username': instance.username,
            'email': instance.email
        }



# Generated by Django 3.1.2 on 2020-10-13 22:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rates', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rate',
            name='base_currency',
            field=models.CharField(db_index=True, default='EUR', max_length=3),
        ),
        migrations.AlterField(
            model_name='rate',
            name='currency',
            field=models.CharField(db_index=True, max_length=3),
        ),
    ]



"""
CurrencyLayer service
"""

import requests
from datetime import date
from django.conf import settings
from typing import Iterator

from . import RatesNotAvailableError, RateService
from ..settings import *

CURRENCYLAYER_API_URL = 'http://api.currencylayer.com/'
CURRENCYLAYER_CURRENCIES_ENDPOINT = 'list'
CURRENCYLAYER_LIVE_ENDPOINT = 'live'
CURRENCYLAYER_HISTORICAL_ENDPOINT = 'historical'
CURRENCYLAYER_TIMEFRAME_ENDPOINT = 'timeframe'

try:
    CURRENCYLAYER_API_KEY = settings.CURRENCYLAYER_API_KEY
except AttributeError:
    pass


class CurrencyLayerService(RateService):
    """
    Currency layer service class
    """

    def available_currencies(self) -> Iterator:
        """
        List availbale currencies for the service
        """
        data = {
            'access_key': CURRENCYLAYER_API_KEY
        }
        url = CURRENCYLAYER_API_URL + CURRENCYLAYER_CURRENCIES_ENDPOINT
        response = requests.get(url=url, data=data)
        if response.status_code == 200:
            data = response.json()
            if 'currencies' in data:
                return data.get('currencies').keys()
            else:
                return []
        else:
            raise RatesNotAvailableError(response.text)

    def fetch_rates(self,
                    base_currency: str = settings.BASE_CURRENCY,
                    currency: str = None,
                    date_obj: date = date.today(),
                    to_obj: date = None) -> []:
        """
        Fetch rates
        :param base_currency: base currency
        :param currency: target currency
        :param date_obj: date of the rate
        :param to_obj: optional range parameter
        """
        data = {
            'access_key': CURRENCYLAYER_API_KEY,
            'source': base_currency
        }
        if currency:
            data['currencies'] = [currency]
        if date_obj == date.today():
            url = CURRENCYLAYER_API_URL + CURRENCYLAYER_LIVE_ENDPOINT
        elif to_obj:
            url = CURRENCYLAYER_API_URL + CURRENCYLAYER_HISTORICAL_ENDPOINT
            data['start_date'] = date_obj
            data['end_date'] = to_obj
        else:
            url = CURRENCYLAYER_API_URL + CURRENCYLAYER_HISTORICAL_ENDPOINT
            data['date'] = date_obj
        response = requests.get(url=url, data=data)
        if response.status_code == 200:
            data = response.json()
            if 'quotes' in data:
                return self.parse_result(base_currency=base_currency, data=data.get('quotes'))
            else:
                return {}
        else:
            raise RatesNotAvailableError(response.text)

    def parse_result(self, base_currency, data: dict) -> []:
        """
        Parse output from currencylayout services
        :param base_currency: base currency
        :param data: dictionnary of data
        """
        output = []
        if 'timeframe' in data:
            # request is timeframe based, so data is agregated by date
            for date_obj, rate_dict in data:
                for code, value in rate_dict:
                    output.append(
                        self.serializer(
                            base_currency=base_currency,
                            currency=code[3:],
                            value=value,
                            date=date_obj)
                    )
        else:
            if 'date' in data:
                # Historical search
                date_obj = data['date']
            else:
                # live search
                date_obj = date.today()
            for code, value in data:
                output.append(
                    self.serializer(
                        base_currency=base_currency,
                        currency=code[3:],
                        value=value,
                        date=date_obj
                    )
                )
        return output
